# -*- coding: utf-8 -*-
"""
Module defining a data format.

A data format has a name, a MIME content-type and is associated
with a reading action for retrieving the data from a file in this
format, a writing action for the opposite purpose and a validating
action to check the data is valid (after reading or before writing).
"""


class DataFormat(object):
    def __init__(self, uid, name, content_type, file_extension, desc):
        self.uid = uid
        self.name = name
        self.content_type = content_type
        self.file_ext = file_extension  # with leading "."
        self.desc = desc
        self.reading_class = None  # ReadAction
        self.writing_class = None  # WriteAction
        self.validating_class = None  # ValidateAction

    def dump(self):
        data = {}
        data[u"formatUid"] = self.uid
        data[u"name"] = self.name
        data[u"desc"] = self.desc
        data[u"contentType"] = self.content_type
        data[u"fileExtension"] = self.file_ext
        return data



# -*- coding: utf-8 -*-
"""
Module containing the functions for transforming all the files contained
in a directory and interacting with the user through the command line.
"""

import os
from os import path as osp
import json
import shutil
import stat
import sys
import gc
from tempfile import mkstemp
import multiprocessing as mp
import logging

from glamconv.transformer.processes import Process
from glamconv import ead

_ACTIONS_REGISTERED = False


def register_actions():
    global _ACTIONS_REGISTERED
    if not _ACTIONS_REGISTERED:
        ead.register()
        _ACTIONS_REGISTERED = True


def build_progress_function(process):
    step_names = [step.action.name for step in process.steps]

    def progress(starting_step_index):
        sys.stdout.write(u"\r\033[K")
        sys.stdout.write(u"    running step: {0} ({1:d}/{2:d})"
                         u"".format(step_names[starting_step_index],
                                    starting_step_index + 1, len(step_names)))
        sys.stdout.flush()
    return progress


def batch_ead2_to_ape(arglist, processes=mp.cpu_count()):
    """convenience entry point to run ead2_to_ape on a list of XML files

    Parameters
    ----------
    arglist: a list of tuple of arguments to be passed to ``ead2_to_ape``,
             (e.g. [('/tmp/input1.xml', '/tmp/output1.xml'),
                    ('tmp/input2.xml', '/tmp/output2.xml')])
    processes: the number of CPU to dispatch the batch on
    """
    pool = mp.Pool(processes=processes)
    pool.map(_convert_to_ape, arglist)


def _convert_to_ape(inputargs):
    try:
        ead2_to_ape(*inputargs)
    except Exception:
        logger = logging.getLogger('glamconv.ead')
        logger.exception('failed to convert_to_ape', inputargs[0])


def ead2_to_ape(input_filepath_or_tree, output_filepath, settings=None):
    """

    Parameters
    -----------

    - `input_filepath_or_tree`: path to the input file or an already parsed
      lxml tree
    - `output_filepath`: path to the output file
    - `settings`: dictionary defining the transformation process (list
      of steps and options).
    """
    if settings is None:
        settings = ead.ead2_to_ape_default_settings()
    register_actions()
    logger = logging.getLogger(settings.get('logger', 'glamconv.ead'))
    if hasattr(input_filepath_or_tree, 'findtext'):
        identifier = input_filepath_or_tree.findtext('eadheader/eadid')
    else:
        identifier = osp.basename(input_filepath_or_tree)
    transform_input = input_filepath_or_tree
    logger.debug('processing %s', identifier)
    transform = Process('ead-2002', 'ape-ead', settings['steps'])
    # Collect files to process
    sys.stdout.write(u"Collecting files to process\n")
    output_dir = osp.dirname(output_filepath)
    if not osp.isdir(output_dir):
        os.makedirs(output_dir)
    if hasattr(input_filepath_or_tree, 'findtext'):
        logger.debug('processing %s',
                     input_filepath_or_tree.findtext('eadheader/eadid'))
    else:
        logger.debug('processing %s', osp.basename(input_filepath_or_tree))
    # Change steps parameters if asked to
# XXX    if adapt_process_params is not None:
# XXX        adapt_process_params(prc, norm_fname)
    # Actual transformation
    temp_fdesc, temp_fname = mkstemp(suffix=transform.output_format.file_ext)
    try:
        with open(temp_fname, "bw") as out_stream:
            log = transform.run(transform_input, out_stream)
        os.close(temp_fdesc)
        # Just to be sure, we do what we can to clean the XML tree
        gc.collect()
        if osp.getsize(temp_fname) > 0:
            shutil.copy(temp_fname, output_filepath)
            os.chmod(output_filepath, (stat.S_IRUSR | stat.S_IWUSR |
                                       stat.S_IRGRP | stat.S_IWGRP |
                                       stat.S_IROTH))
        else:
            log.warning(
                "transformation of %s didn't generate any output",
                transform_input
            )
    finally:
        if osp.isfile(temp_fname):
            os.remove(temp_fname)
    # Log file
    log_fname = osp.join(
        output_dir,
        u"{0}.log.json".format(osp.splitext(osp.basename(output_filepath))[0])
    )
    with open(log_fname, u"w") as log_stream:
        json.dump(log.dump(), log_stream)


def transform_files_from_dir(process_filename, input_dir, output_dir,
                             recursive=True, adapt_process_params=None):
    # Build process
    sys.stdout.write(u"Building transformation process\n")
    with open(process_filename, "br") as inp:
        prc = Process.from_json(json.load(inp))
    prg_func = build_progress_function(prc)
    inp_ext = prc.input_format.file_ext
    out_ext = prc.output_format.file_ext
    # Collect files to process
    sys.stdout.write(u"Collecting files to process\n")
    files_to_process = []
    if recursive:
        files_iter = os.walk(input_dir, followlinks=True)
    else:
        files_iter = [(input_dir, [], fname)
                      for fname in os.listdir(input_dir)]
    for dirname, subdirs, files in files_iter:
        out_dirname = osp.join(output_dir, osp.relpath(dirname, input_dir))
        # Normalize output dirname
        norm_out_dirname = out_dirname.decode(u"ascii", u"replace")
        norm_out_dirname = norm_out_dirname.replace(u"\ufffd", u"X")
        norm_out_dirname = norm_out_dirname.replace(u" ", u"_")
        # Create output structure
        if not osp.isdir(norm_out_dirname):
            os.makedirs(norm_out_dirname)
        for fname in files:
            if osp.splitext(fname)[1] not in (inp_ext, inp_ext.upper()):
                continue
            # Normalize filename
            norm_fname = fname.decode(u"ascii", u"replace")
            norm_fname = norm_fname.replace(u"\ufffd", u"X")
            norm_fname = norm_fname.replace(u" ", u"_")
            files_to_process.append(
                (norm_fname, dirname, fname, norm_out_dirname))
    files_number = len(files_to_process)
    sys.stdout.write(u"    found {0:d} files to process\n\n"
                     u"".format(files_number))
    # Run transformation
    abstract = {u"OK": [], u"OK-invalid-input": 0, u"INVALID": [],
                u"FAILED": [], u"ERROR": []}
    for file_idx, (norm_fname, inp_dirname, inp_filename, out_dirname) in \
                                                enumerate(files_to_process):  # noqa
        sys.stdout.write(u"*** {0} ({1:d}/{2:d})\n"
                         u"".format(norm_fname, file_idx, files_number))
        inp_fname = osp.join(inp_dirname, inp_filename)
        # Change steps parameters if asked to
        if adapt_process_params is not None:
            adapt_process_params(prc, norm_fname)
        # Actual transformation
        temp_fdesc, temp_fname = mkstemp(suffix=out_ext)
        with open(temp_fname, "bw") as out_stream:
            with open(inp_fname, "br") as inp_stream:
                log = prc.run(inp_stream, out_stream, prg_func)
        os.close(temp_fdesc)
        sys.stdout.write(u"\r\033[K")
        sys.stdout.flush()
        # Just to be sure, we do what we can to clean the XML tree
        gc.collect()
        # Choose output file and copy it in output dir
        fname_name, inp_ext = osp.splitext(norm_fname)
        out_fname = osp.join(out_dirname, fname_name)
        if log.output_validation:
            # Result is valid
            abstract[u"OK"].append(norm_fname)
            out_fname += out_ext
            msg = u"    OK"
            if not log.input_validation:
                msg += u" even if the input file was not valid"
                abstract[u"OK-invalid-input"] += 1
        elif log.failure:
            # An exception occured
            abstract[u"ERROR"].append(
                u"{0} (in step #{1:d})".format(norm_fname,
                                               (log.failure_step_index + 1)))
            out_fname += u".ERROR{0}".format(out_ext)
            msg = (u"    ERROR during transformation\n{0}"
                   u"".format(log.failure_message))
        elif log.input_validation:
            # Result is not valid and input was valid
            abstract[u"FAILED"].append(norm_fname)
            out_fname += u".FAILED{0}".format(out_ext)
            msg = (u"    FAILED\n{0}"
                   u"".format(log.output_validation_message))
        else:
            # Result is not valid but input wasn't valid
            abstract[u"INVALID"].append(norm_fname)
            out_fname += u".INVALID{0}".format(out_ext)
            msg = (u"    INVALID input\n{0}\n    INVALID output\n{1}"
                   u"".format(log.input_validation_message,
                              log.output_validation_message))
        sys.stdout.write(u"{0}\n".format(msg))
        if osp.getsize(temp_fname) > 0:
            shutil.copy(temp_fname, out_fname)
            os.chmod(out_fname, (stat.S_IRUSR | stat.S_IWUSR |
                                 stat.S_IRGRP | stat.S_IWGRP |
                                 stat.S_IROTH))
        else:
            os.remove(temp_fname)
            sys.stdout.write(u"    no output generated\n")
        # Log file
        log_fname = osp.join(out_dirname,
                             u"{0}.log.json".format(fname_name))
        with open(log_fname, u"bw") as log_stream:
            json.dump(log.dump(), log_stream)
    sys.stdout.write(u"\nSynthesis:\n")
    sys.stdout.write(u"  Files to process: {0:d}\n".format(files_number))
    for key in (u"INVALID", u"FAILED", u"ERROR", u"OK"):
        sys.stdout.write(u"  {0}: {1:d} files\n".format(key,
                                                        len(abstract[key])))
        if key != u"OK" and len(abstract[key]) > 0:
            sys.stdout.write(u" ".join(abstract[key]))
            sys.stdout.write(u"\n")
        if key == u"OK" and abstract[u"OK-invalid-input"] > 0:
            sys.stdout.write(
                u"      including {0:d} files with invalid input\n"
                u"".format(abstract[u"OK-invalid-input"]))



# -*- coding: utf-8 -*-
"""
Module containing actions that modify the namespace of the XML elements.
"""

from lxml import etree
from glamconv.ead.utils import NS, split_qname
from glamconv.ead.formats import EAD_2002, EAD_APE
from glamconv.transformer.actions import TransformAction


class EadNamespaceAdder(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"ead-namespace-adder"
    name = u"Adding the Ape-EAD namespace"
    category = u"Naming"
    desc = (u"This action puts all the XML elements with an empty namespace "
            u"inside the Ape-EAD standard namespace. The attribute names are "
            u"not changed. All the other actions designed for EAD-2002 "
            u"documents expect the elements to have an empty namespace "
            u"(such as the elements in EAD 2002). Therefore, be sure to run "
            u"this action at the end of the transformation, just before the "
            u"Ape-EAD validation.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        for elt in xml_root.iter(etree.Element):
            namespace, local_name = split_qname(elt.tag)
            if namespace is None:
                elt.tag = u"{{{0}}}{1}".format(NS["ead"], local_name)
                count += 1
        if count > 0:
            logger.info(u"{0:d} elements with an empty namespace have been "
                        u"transfered in Ape-EAD namespace".format(count))
        return xml_root


class EadNamespaceRemover(TransformAction):
    applicable_for = (EAD_2002, EAD_APE)
    uid = u"ead-namespace-remover"
    name = u"Removing the Ape-EAD namespace"
    category = u"Naming"
    desc = (u"This action puts all the XML elements with the Ape-EAD standard "
            u"namespace inside the empty namespace. The attribute names are "
            u"not changed. This action does the opposite of the action that "
            u"adds the Ape-EAD namespace to the XML elements.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        for elt in xml_root.iter(etree.Element):
            namespace, local_name = split_qname(elt.tag)
            if namespace == NS[u"ead"]:
                elt.tag = u"{{}}{0}".format(local_name)
                count += 1
        if count > 0:
            logger.info(u"{0:d} elements with the Ape-EAD namespace have been "
                        u"transfered in the empty namespace".format(count))
        return xml_root



# -*- coding: utf-8 -*-
"""
Module defining several useful functions that will be used inside
the transform actions defined in the other modules of this sub-package.
"""

import re


NS = {
    'ead': 'urn:isbn:1-931666-22-9',
    'xlink': 'http://www.w3.org/1999/xlink',
}


QNAME_PATTERN = re.compile("({.*})?(.+)")


def split_qname(qname):
    namespace, local_name = QNAME_PATTERN.match(qname).groups()
    if namespace is not None:
        namespace = namespace[1:-1]
    if namespace == u"":
        namespace = None
    return namespace, local_name


def write_hierarchy(xml_elt):
    ancest = xml_elt.xpath("ancestor-or-self::*")
    hierarchy = u""
    for idx, elt in enumerate(ancest):
        hierarchy += "/{0}".format(elt.xpath("name()"))
        if elt.get(u"id"):
            hierarchy += u'[id="{0}"]'.format(elt.get(u"id"))
        elif idx != 0:
            xpath_expr = (u'count(preceding-sibling::{0})+1'
                          u''.format(elt.xpath("name()")))
            elt_index = int(elt.xpath(xpath_expr, namespaces=NS))
            xpath_expr = u'count(../{0})'.format(elt.xpath("name()"))
            max_index = int(elt.xpath(xpath_expr, namespaces=NS))
            if max_index > 1:
                hierarchy += u'[{0}]'.format(elt_index)
    return hierarchy


def log_element(xml_elt, attributes=None, text_content=False, msg=u""):
    log_entry = write_hierarchy(xml_elt)
    if msg:
        log_entry += u"\n    " + msg
    if attributes:
        log_entry += u"\n   "
        for attrname in attributes:
            attrval = xml_elt.xpath(u"string(@{0})".format(attrname),
                                    namespaces=NS)
            if len(attrval) == 0:
                continue
            if len(attrval) > 50:
                attrval = attrval[:47] + u"..."
            log_entry += u" {0}=\"{1}\"".format(attrname, attrval)
    if text_content:
        log_entry += u"\n    Text content: "
        txt = xml_elt.xpath(u"string()")
        if len(txt) > 50:
            txt = txt[:47] + u"..."
        log_entry += txt
    return log_entry


def suppress_element(xml_elt):
    parent = xml_elt.getparent()
    text = (xml_elt.text or u"") + (xml_elt.tail or u"")
    previous = xml_elt.getprevious()
    if previous is not None:
        previous.tail = (previous.tail or u"") + text
    else:
        parent.text = (parent.text or u"") + text
    parent.remove(xml_elt)


def add_text_around_element(xml_elt, before=u"", after=u""):
    previous = xml_elt.getprevious()
    parent = xml_elt.getparent()
    if before:
        if previous is not None:
            previous.tail = (previous.tail or u"") + before
        else:
            parent.text = (parent.text or u"") + before
    if after:
        xml_elt.tail = (xml_elt.tail or u"") + after


def adjust_content_in_mixed(xml_elt, keep_tags=tuple()):
    """
    Erase from ``xml_elt`` all the XML elements except those whose tag name
    is given in ``keep_tags`` tuple. The text inside ``xml_elt`` and its
    children is kept (even if the elements are removed).

    If an XML element must be kept, it is kept as it is (with all its
    sub-elements). Therefore, if you want to erase XML elements that can
    occur at several levels in the XML tree, you have to walk depth-first in
    the tree and call the function first from the bottom levels and upwards.
    """
    for child in xml_elt:
        if child.tag in keep_tags:
            # If child must be kept, keep it as it is and go to next child
            continue
        # Recursively call the same function to get all the text content from
        # child (``keep_tags`` is empty because we don't want to keep any
        # XML element from ``child`` that will not be kept).
        adjust_content_in_mixed(child, tuple())
        # Keep text content of ``child`` inside the element before ``child``
        child_text_content = (child.text or u"") + (child.tail or u"")
        if child_text_content:
            previous = child.getprevious()
            if previous is not None:
                previous.tail = (previous.tail or u"") + child_text_content
            else:
                xml_elt.text = (xml_elt.text or u"") + child_text_content
        # Finally, remove child from the XML tree
        xml_elt.remove(child)


def insert_child_at_element_beginning(xml_elt, child):
    if xml_elt.text is not None:
        child.tail = (child.tail or u"") + xml_elt.text
        xml_elt.text = None
    xml_elt.insert(0, child)


def insert_text_at_element_end(xml_elt, text):
    if not text:
        return
    if len(xml_elt) == 0:
        xml_elt.text = (xml_elt.text or u"") + text
    else:
        last_child = xml_elt[-1]
        last_child.tail = (last_child.tail or u"") + text


def insert_text_before_element(xml_elt, text):
    if not text:
        return
    prev = xml_elt.getprevious()
    if prev is None:
        parent = xml_elt.get_parent()
        parent.text = (parent.text or u"") + text
    else:
        prev.tail = (prev.tail or u"") + text



# -*- coding: utf-8 -*-
"""
Module containing actions for changing the values of the vocabulary-controlled
attributes in order to have a legit value.
"""

import re
from glamconv.ead.utils import log_element, add_text_around_element
from glamconv.ead.formats import EAD_2002
from glamconv.transformer.actions import TransformAction


OTHLEV_REGEXP = re.compile(r"^([A-Za-z]|:|_)([0-9A-Za-z]|:|_|-|\.)*$")


class ArchdescLevelAttribAdjuster(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"archdesc-level-adjuster"
    name = u"Adjusting the value of level attribute in <archdesc>"
    category = u"Cleansing"
    desc = (u"In the 'level' attribute of <archdesc> elements the only "
            u"possible legit value is \"fonds\". This action sets 'level' "
            u"to its legit value and puts its previous value in the "
            u"'otherlevel' attribute of <archsdesc> if it is empty.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for arch in xml_root.xpath(
                u'.//archdesc[not(@level) or @level!="fonds"]'):
            msg = u""
            value = arch.get(u"level", u"")
            if value not in [u"", u"otherlevel"] \
               and arch.get(u"otherlevel", u"") != u"" \
               and OTHLEV_REGEXP.match(value) is not None:
                arch.set(u"otherlevel", value)
            else:
                msg = (u"    Old value: {0}".format(value))
            arch.set(u"level", u"fonds")
            count += 1
            if log_details:
                log_data.append(log_element(arch, msg=msg))
        if count > 0:
            logger.warning(
                u"The 'level' attribute in the <archdesc> element can only "
                u"have one value: \"fonds\". {0:d} elements have been "
                u"corrected.".format(count))
            if log_details:
                logger.warning(u"The following elements have been corrected:",
                               u"\n".join(log_data))
        return xml_root


class DscTypeAttribAdjuster(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"dsc-type-adjuster"
    name = u"Adjusting the value of type attribute in <dsc>"
    category = u"Cleansing"
    desc = (u"In the 'type' attribute of <dsc> elements the only possible "
            u"legit value is \"othertype\". This action sets 'type' "
            u"to its legit value.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for dsc in xml_root.xpath(u'.//dsc[@type!="othertype"]'):
            if log_details:
                msg = u"    Old value: {0}".format(dsc.get("type"))
                log_data.append(log_element(dsc, msg=msg))
            dsc.attrib['type'] = 'othertype'
            count += 1
        if count > 0:
            logger.warning(
                u"The 'type' attribute in the <dsc> element can only have "
                u"one value: \"othertype\". {0:d} elements have been "
                u"corrected.".format(count))
            if log_details:
                logger.warning(u"The following elements have been corrected:",
                               u"\n".join(log_data))
        return xml_root


class EmphRenderAttribAdjuster(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"emph-render-adjuster"
    name = u"Adjusting the value of render attribute in <emph>"
    category = u"Cleansing"
    desc = (u"In the 'render' attribute of <emph> elements the only possible "
            u"legit values are \"bold\" and \"italic\". If the value contains "
            u"\"bold\", sets the value to \"bold\", else if the value "
            u"contains \"italic\", sets the value to \"italic\", else deletes "
            u"the value. If the value contains the mention of quotes (single "
            u"or double), adds them to the text before and after the <emph> "
            u"element.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for emph in xml_root.xpath(
                u'.//emph[@render!="bold" and @render!="italic"]'):
            value = emph.attrib.pop("render", u"")
            if value == u"":
                continue
            if u"bold" in value:
                emph.set(u"render", u"bold")
            elif u"italic" in value:
                emph.set(u"render", u"italic")
            if u"doublequote" in value:
                add_text_around_element(emph, u"« ", u" »")
            elif u"singlequote" in value:
                add_text_around_element(emph, u"\u2018", u"\u2019")
            if log_details:
                msg = u"    Old value: {0}".format(value)
                log_data.append(log_element(emph, msg=msg, text_content=True))
            count += 1
        if count > 0:
            logger.warning(
                u"The render attribute in the <emph> element can only have "
                u"the values \"bold\" or \"italic\". {0:d} elements have been "
                u"corrected.".format(count))
            if log_details:
                logger.warning(u"The following elements have been corrected:"
                               u"\n".join(log_data))
        return xml_root


LNGS = (u"roh sco scn rom ron alg oss ale alb scc mni nld mno osa mnc scr oci "
        u"mwr crp lua jav hrv lub lun twi aus roa ven uga mwl ger mga hit fas "
        u"ssw lug luo fat fan fao lui tgl ita geo him hin mlg din mun hye gba "
        u"guj cmc slk bad nep iba ban ady div bam bak shn bai arp tel tem arw "
        u"nwc ara tkl arc nbl ter sel tet arm arn ave lus mus sun kut suk kur "
        u"wol kum sus iro new oji kua sux mlt iku hai tup tur men tut mic grc "
        u"tuk tum mul lez nav cop cos cor gla bos gwi gle eka glg akk und dra "
        u"aka tlh bod glv jrb vie ipk por uzb pon pol sah sgn sga tgk bre apa "
        u"wak bra aym cha chb che fre chg swa chi chk fro chm chn cho chp cpf "
        u"chr xal chu chv chy fry dut hat msa gay oto ota hmn hmo ice niu myv "
        u"iii gaa fur ndo ibo ina car slv xho deu cau cat cai slo del fil den "
        u"ilo inh ful sla cad tat jpn vol myn vot ind dzo spa tam jpr tah tai "
        u"bis cze pap afh pau eng afa ewe csb phi paa nyn nyo nym bnt nya sio "
        u"sin afr map mas mar lah phn sna may kir yao snk dgr syr mac mad mag "
        u"mai mah lav mal mao man egy pag znd sit gmh epo jbo tiv tha dum fon "
        u"zen kbd enm kha sam tsn tso dsb cus ell fiu ssa wen wel byn elx gem "
        u"uig ada fij tli fin hau eus haw bin amh non ceb bih mos dan nog bat "
        u"nob dak moh ces mon dar mol son day nor bas cel pal baq pam hil kpe "
        u"dua sot lad lam mdf snd tvl lao abk mdr ijo gon goh sms per smo smn "
        u"peo smj smi pan tmh got sme bla sma gor hsb nic nia run ast mkd sag "
        u"cre fra bik orm que ori rus crh asm pus kik gez srd ltz ach nde sqi "
        u"ath kru srr ido srp nub kro wln isl ava krc nds zun zul kin umb sog "
        u"nso swe som yap lat frm art mak zap yid kok vai kom her kon ukr lol "
        u"mkh ton heb loz kos kor was tog ira pli sid bur hun hup bul wal bua "
        u"bug cym udm bej gil ben bel bem tsi aze war zha ace rum aar ber nzi "
        u"zho nno sai ang pra san bho sal pro arg raj sad khi rar khm rap kho "
        u"sas ine sem sat min lim lin nai tir nah lit efi nap gre grb btk nau "
        u"grn ypk mis tig yor tib kac kab kaa ile kan kam kal kas kar mya est "
        u"kaw kau kat kaz tyv awa kmb urd doi ewo cpe tpi mri dyu cpp bal inc")
LNGS = tuple(LNGS.split(" "))


class LanguageLangcodeAttribAdjuster(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"language-langcode-adjuster"
    name = u"Adjusting the value of langcode attribute in <language>"
    category = u"Cleansing"
    desc = (
        u"The 'langcode' attribute of <language> elements can only take "
        u"a definite set of values. This action erases the attributes that "
        u"don't have a legit value.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for elt in xml_root.xpath(u'.//language[@langcode]'):
            value = elt.attrib.pop(u"langcode", u"").lower()
            if value in LNGS:
                elt.set(u"langcode", value)
            else:
                if log_details:
                    msg = u"    Deleted value: {0}".format(value)
                    log_data.append(log_element(elt, msg=msg))
                count += 1
        if count > 0:
            logger.warning(
                u"The 'langcode' attribute in the <language> element can only "
                u"have a value taken from a fixed vocabulary. {0:d} elements "
                u"have had their 'langcode' attribute deleted because of a "
                u"non-legit value.".format(count))
            if log_details:
                logger.warning(u"The following elements have been corrected:"
                               u"\n".join(log_data))
        return xml_root


SCRS = (u"Tfng Shaw Cirt Guru Hebr Geor Zzzz Phnx Hrkt Plrd Ogam Telu Bopo "
        u"Dsrt Xsux Visp Hmng Hano Bali Gujr Hang Thaa Sinh Hans Hant Talu "
        u"Mong Deva Sara Qabx Tagb Inds Xpeo Geok Tale Mymr Mand Perm Bugi "
        u"Phag Brai Brah Mlym Tibt Kali Batk Vaii Sylo Lina Teng Mero Limb "
        u"Kana Yiii Roro Java Taml Orya Laoo Ugar Cyrl Nkoo Armn Cyrs Latg "
        u"Latf Khmr Khar Egyh Latn Maya Ethi Goth Ital Arab Zxxx Buhd Copt "
        u"Thai Cprt Linb Lepc Osma Runr Glag Hira Syre Hani Orkh Hung Grek "
        u"Qaaa Egyd Cher Zyyy Cham Syrc Blis Cans Beng Egyp Syrj Tglg Syrn "
        u"Knda")
SCRS = tuple(SCRS.split(" "))


class LanguageScriptcodeAttribAdjuster(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"language-scriptcode-adjuster"
    name = u"Adjusting the value of scriptcode attribute in <language>"
    category = u"Cleansing"
    desc = (
        u"The 'scriptcode' attribute of <language> elements can only take "
        u"a definite set of values. This action erases the attributes that "
        u"don't have a legit value.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for elt in xml_root.xpath(u'.//language[@scriptcode]'):
            value = elt.attrib.pop(u"scriptcode", u"").capitalize()
            if value in SCRS:
                elt.set(u"scriptcode", value)
            else:
                if log_details:
                    msg = u"    Deleted value: {0}".format(value)
                    log_data.append(log_element(elt, msg=msg))
                count += 1
        if count > 0:
            logger.warning(
                u"The 'scriptcode' attribute in the <language> element can "
                u"only have a value taken from a fixed vocabulary. {0:d} "
                u"elements have had their 'scriptcode' attribute deleted "
                u"because of a non-legit value.".format(count))
            if log_details:
                logger.warning(u"The following elements have been corrected:"
                               u"\n".join(log_data))
        # Return the cleaned EAD tree
        return xml_root


class ListAttribsAdjuster(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"list-attribs-adjuster"
    name = u"Adjusting the value of attributes in <list>"
    category = u"Cleansing"
    desc = (u"The 'numeration' and the 'type' attributes of <list> elements "
            u"can only have a few legit value. This action erases the "
            u"attributes whose value is not legit.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for list_elt in xml_root.xpath(u'.//list[@numeration or @type]'):
            msg = u""
            if list_elt.get(u"numeration") not in [None, u"arabic"]:
                val = list_elt.attrib.pop(u"numeration")
                msg += u" numeration= \"{0}\"".format(val)
            if list_elt.get(u"type") not in [None, u"ordered", u"marked"]:
                val = list_elt.attrib.pop(u"type")
                msg += u" type= \"{0}\"".format(val)
            if msg != "":
                if log_details:
                    log_data.append(
                        log_element(list_elt,
                                    msg=(u"    Deleted attributes:"+msg)))
                count += 1
        if count > 0:
            logger.warning(
                u"The attributes of <list> elements can only have specific "
                u"values. {0:d} elements have had their attributes deleted "
                u"because of a non-legit value:".format(count)),
            if log_details:
                logger.warning(u"The following elements have been corrected:"
                               u"\n".join(log_data))
        return xml_root



# -*- coding: utf-8 -*-
"""
Module defining classes for describing the parameters of the actions.

The parameter classes contain the definition of the parameter (name,
description, type) but also the mechanism to convert the specified
values for the parameter (most of the time these are text values read
from a file or received from a socket) into actual values that can be
used by the action (numbers, dates, etc.)
"""

SINGLE_TEXT_FIELD = u"single-text-field"
"""
Form type for the parameters whose value is a single text value.

This type can be used in the GUI to correctly display the form where
the user gives a value for the parameter.

Type: :class:`str`
"""
MULTIPLE_TEXT_COUPLES_FIELD = u"multiple-text-couples-field"
"""
Form type for the parameters whose value is a list of couples of text values.

This type can be used in the GUI to correctly display the form where
the user gives a value for the parameter.

Type: :class:`str`
"""


class AbstractParameter(object):
    form_type = None

    def __init__(self, uid, name, desc, type_desc):
        self.uid = uid
        self.name = name
        self.desc = desc
        self.type_desc = type_desc
        if self.__class__ is AbstractParameter:
            raise NotImplementedError(u"Abstract class can't be implemented")

    def _get_def_value(self):
        raise NotImplementedError(u"Abstract method can't be called")

    def _convert(self, specif_val):
        raise NotImplementedError(u"Abstract method can't be called")

    def dump(self):
        return {u"uid": self.uid,
                u"name": self.name,
                u"desc": self.desc,
                u"typeDesc": self.type_desc,
                u"formType": self.form_type}

    def build_value(self, specif_val):
        if specif_val is None:
            return self._get_def_value()
        else:
            return self._convert(specif_val)

    def log_value(self, value):
        return (self.name, u"{0}".format(value))


class SingleParameter(AbstractParameter):
    form_type = SINGLE_TEXT_FIELD

    def __init__(self, uid, name, desc, type_desc,
                 type_conv, def_value=None):
        super(SingleParameter, self).__init__(uid, name, desc, type_desc)
        self._type_conv = type_conv
        self._def_value = def_value

    def _get_def_value(self):
        return self._def_value

    def _convert(self, specif_val):
        try:
            value = self._type_conv(specif_val)
        except ValueError:
            raise ValueError(u"Can't convert to target type with {0} from: {1}"
                             u"".format(self._type_conv, specif_val))
        return value


class CouplesParameter(AbstractParameter):
    form_type = MULTIPLE_TEXT_COUPLES_FIELD

    def __init__(self, uid, name, desc, type_desc,
                 type_conv_1, type_conv_2, def_value=None):
        super(CouplesParameter, self).__init__(uid, name, desc, type_desc)
        self._type_conv_1 = type_conv_1
        self._type_conv_2 = type_conv_2
        self._def_value = def_value

    def _get_def_value(self):
        if isinstance(self._def_value, (tuple, list)):
            return list(self._def_value)[:]
        else:
            return self._def_value

    def _convert(self, specif_val):
        couples = []
        if not isinstance(specif_val, (tuple, list)):
            raise ValueError(u"Expects a list of couples and got: {0}"
                             u"".format(specif_val))
        for elts in specif_val:
            try:
                elt_1, elt_2 = elts
            except ValueError:
                raise ValueError(u"Expects a list of couples and got: {0}"
                                 u"".format(specif_val))
            try:
                val_1 = self._type_conv_1(elt_1)
            except ValueError:
                raise ValueError(
                    u"Can't convert to target type with {0} from: "
                    u"{1}".format(self._type_conv_1, elt_1))
            try:
                val_2 = self._type_conv_2(elt_2)
            except ValueError:
                raise ValueError(
                    u"Can't convert to target type with {0} from: "
                    u"{1}".format(self._type_conv_2, elt_2))
            couples.append((val_1, val_2))
        return couples

    def log_value(self, value):
        return (self.name, u"\n".join([u"{0}: {1}".format(elt_1, elt_2)
                                       for (elt_1, elt_2) in value]))



# -*- coding: utf-8 -*-
"""
Module defining two global libraries that contains all the known formats
and the known actions.

The :func:`~glamconv.transformer.libraries.register_action` and
:func:`~glamconv.transformer.libraries.register_format` functions can
be used to register an action class or a format inside these libraries. The
:func:`~glamconv.transformer.libraries.get_action` and
:func:`~glamconv.transformer.libraries.get_format` functions can be used
to retrieve an action class or a format from these libraries. A format is
registred with its unique identifier (``uid``), an action is registred with
the identifier of the format it applies to and its name (if the action can be
applied to multiple formats, it is registred multiple times).
"""

import operator
from glamconv.transformer.actions import AbstractAction
from glamconv.transformer.formats import DataFormat


ACTIONS_LIBRARY = {}  # {format_uid: {action_uid: AbstractAction, }, }
FORMATS_LIBRARY = {}  # {format_uid: DataFormat(), }


# Actions

def register_action(action_class, format_uid=None):
    assert issubclass(action_class, AbstractAction)
    if format_uid is not None:
        if format_uid not in ACTIONS_LIBRARY:
            ACTIONS_LIBRARY[format_uid] = {}
        ACTIONS_LIBRARY[format_uid][action_class.uid] = action_class
    else:
        for frmt in action_class.applicable_for:
            if format_uid not in ACTIONS_LIBRARY:
                ACTIONS_LIBRARY[format_uid] = {}
            ACTIONS_LIBRARY[format_uid][action_class.uid] = action_class


def get_action(format_uid, action_uid):
    return ACTIONS_LIBRARY[format_uid][action_uid]


def list_action_categories(format_uid, family=None):
    sel_actions = ACTIONS_LIBRARY.get(format_uid, {}).values()
    if family is not None:
        sel_actions = [act for act in sel_actions if act.family == family]
    return sorted(set([act.category for act in sel_actions]))


def list_actions(format_uid, family=None, category=None):
    sel_actions = ACTIONS_LIBRARY.get(format_uid, {}).values()
    if family is not None:
        sel_actions = [act for act in sel_actions if act.family == family]
    if category is not None:
        sel_actions = [act for act in sel_actions if act.category == category]
    return sorted(sel_actions, key=operator.attrgetter(u"name"))


# Formats
def register_format(data_format):
    assert isinstance(data_format, DataFormat)
    FORMATS_LIBRARY[data_format.uid] = data_format
    register_action(data_format.reading_class, data_format.uid)
    register_action(data_format.writing_class, data_format.uid)
    register_action(data_format.validating_class, data_format.uid)


def get_format(format_uid):
    return FORMATS_LIBRARY[format_uid]


def list_formats():
    return sorted(FORMATS_LIBRARY.values(), key=operator.attrgetter(u"name"))



# -*- coding: utf-8 -*-
"""
Module containing actions for deleting attributes or sub-elements from the XML
elements because they don't exist anymore in Ape-EAD format.
"""

from lxml import etree
from glamconv.ead.utils import log_element, split_qname
from glamconv.ead.formats import EAD_2002
from glamconv.transformer.actions import TransformAction


ATTR_TO_ERASE = {
    u"abbr": (u"altrender", u"audience", u"id",),
    u"abstract": (u"altrender", u"audience", u"id", u"langcode",),
    u"accessrestrict": (u"altrender", u"audience", u"id", u"type",),
    u"accruals": (u"altrender", u"audience", u"id",),
    u"acqinfo": (u"altrender", u"audience", u"id",),
    u"address": (u"altrender", u"audience", u"id",),
    u"addressline": (u"altrender", u"audience", u"id",),
    u"altformavail": (u"altrender", u"audience", u"id", u"type",),
    u"appraisal": (u"altrender", u"audience", u"id",),
    u"archdesc": (u"altrender", u"audience", u"id",),
    u"arrangement": (u"altrender", u"audience", u"id",),
    u"author": (u"altrender", u"audience", u"id",),
    u"bibliography": (u"altrender", u"audience", u"id",),
    u"bibref": (u"altrender", u"audience", u"encodinganalog", u"entityref",
                u"id", u"linktype",),
    u"bioghist": (u"altrender", u"audience", u"id",),
    u"blockquote": (u"altrender", u"audience", u"id",),
    u"c": (u"altrender", u"tpattern",),
    u"c01": (u"altrender", u"tpattern",),
    u"c02": (u"altrender", u"tpattern",),
    u"c03": (u"altrender", u"tpattern",),
    u"c04": (u"altrender", u"tpattern",),
    u"c05": (u"altrender", u"tpattern",),
    u"c06": (u"altrender", u"tpattern",),
    u"c07": (u"altrender", u"tpattern",),
    u"c08": (u"altrender", u"tpattern",),
    u"c09": (u"altrender", u"tpattern",),
    u"c10": (u"altrender", u"tpattern",),
    u"c11": (u"altrender", u"tpattern",),
    u"c12": (u"altrender", u"tpattern",),
    u"change": (u"altrender",),
    u"colspec": (u"align", u"char", u"charoff", u"colspec", u"colwidth",
                 u"rowsep",),
    u"container": (u"altrender", u"audience", u"encodinganalog", u"id",
                   u"label",),
    u"controlaccess": (u"altrender", u"audience", u"encodinganalog", u"id",),
    u"corpname": (u"altrender", u"audience", u"encodinganalog", u"id",
                  u"normal", u"role", u"rules", u"source"),
    u"creation": (u"altrender", u"audience", u"encodinganalog", u"id",),
    u"custodhist": (u"altrender", u"audience", u"id",),
    u"dao": (u"altrender", u"audience", u"entityref", u"id", u"linktype",),
    u"daoloc": (u"altrender", u"audience", u"entityref", u"id", u"linktype",),
    u"date": (u"altrender", u"audience", u"certainty", u"id", u"type",),
    u"descrules": (u"altrender",),
    u"did": (u"altrender", u"audience", u"encodinganalog", u"id",),
    u"dimensions": (u"altrender", u"audience", u"encodinganalog", u"id",
                    u"label",),
    u"dsc": (u"altrender", u"audience", u"encodinganalog", u"id", u"othertype",
             u"label", u"tpattern",),
    u"ead": (u"altrender", u"relatedencoding",),
    u"eadheader": (u"altrender", u"audience", u"encodinganalog",
                   u"findaidstatus", u"id",),
    u"eadid": (u"encodinganalog", u"publicid", u"urn",),
    u"emph": (u"altrender", u"id",),
    u"entry": (u"align", u"altrender", u"audience", u"char", u"charoff",
               u"colname", u"colsep", u"id", u"morerows", u"nameend",
               u"namest", u"rowsep", u"valign",),
    u"expan": (u"altrender", u"audience", u"id",),
    u"extent": (u"altrender", u"audience", u"encodinganalog", u"id",
                u"label", u"type",),
    u"extptr": (u"altrender", u"audience", u"entityref", u"id", u"linktype",),
    u"extref": (u"altrender", u"audience", u"entityref", u"id", u"linktype",),
    u"famname": (u"altrender", u"audience", u"encodinganalog", u"id",
                 u"normal", u"role", u"rules", u"source",),
    u"filedesc": (u"altrender", u"audience", u"encodinganalog", u"id",),
    u"fileplan": (u"altrender", u"audience", u"encodinganalog", u"id",),
    u"function": (u"altrender", u"audience", u"authfilenumber",
                  u"encodinganalog", u"id", u"normal", u"rules", u"source",),
    u"genreform": (u"altrender", u"audience", u"authfilenumber",
                   u"encodinganalog", u"id", u"normal", u"rules", u"source",
                   u"type",),
    u"geogname": (u"altrender", u"audience", u"authfilenumber",
                  u"encodinganalog", u"id", u"normal", u"role", u"rules",
                  u"source",),
    u"head": (u"althead", u"altrender", u"audience", u"id",),
    u"imprint": (u"altrender", u"audience", u"encodinganalog", u"id",),
    u"item": (u"altrender", u"audience", u"id",),
    u"label": (u"altrender", u"audience", u"id",),
    u"langmaterial": (u"altrender", u"audience", u"id", u"label",),
    u"language": (u"altrender", u"audience", u"id",),
    u"langusage": (u"altrender", u"audience", u"encodinganalog", u"id",),
    u"list": (u"altrender", u"audience", u"continuation", u"id", u"mark",),
    u"materialspec": (u"altrender", u"audience", u"encodinganalog", u"id",
                      u"label", u"type",),
    u"name": (u"altrender", u"audience", u"encodinganalog", u"id", u"normal",
              u"role", u"rules", u"source",),
    u"note": (u"actuate", u"altrender", u"audience", u"id", u"show",),
    u"occupation": (u"altrender", u"audience", u"authfilenumber",
                    u"encodinganalog", u"id", u"normal", u"rules", u"source",),
    u"odd": (u"altrender", u"audience", u"id", u"type",),
    u"originalsloc": (u"altrender", u"audience", u"id", u"type",),
    u"origination": (u"altrender", u"audience", u"id",),
    u"otherfindaid": (u"altrender", u"audience", u"id",),
    u"p": (u"altrender", u"audience", u"id",),
    u"persname": (u"altrender", u"audience", u"encodinganalog", u"id",
                  u"normal", u"role", u"rules", u"source"),
    u"physdesc": (u"altrender", u"audience", u"id", u"label", u"rules",
                  u"source",),
    u"physfacet": (u"altrender", u"audience", u"encodinganalog", u"id",
                   u"label", u"rules", u"source", u"unit",),
    u"physloc": (u"altrender", u"audience", u"encodinganalog", u"id",
                 u"parent", u"type",),
    u"phystech": (u"altrender", u"audience", u"id", u"type",),
    u"prefercite": (u"altrender", u"audience", u"id",),
    u"processinfo": (u"altrender", u"audience", u"id", u"type",),
    u"profiledesc": (u"altrender", u"audience", u"encodinganalog", u"id",),
    u"ptr": (u"altrender", u"audience", u"id", u"linktype",),
    u"publicationstmt": (u"altrender", u"audience", u"encodinganalog", u"id",),
    u"publisher": (u"altrender", u"audience", u"id",),
    u"ref": (u"altrender", u"audience", u"id", u"linktype",),
    u"relatedmaterial": (u"altrender", u"audience", u"id", u"type",),
    u"repository": (u"altrender", u"audience", u"encodinganalog", u"id",
                    u"label",),
    u"revisiondesc": (u"altrender",),
    u"row": (u"altrender", u"audience", u"id", u"rowsep", u"valign",),
    u"scopecontent": (u"altrender", u"audience", u"id",),
    u"separatedmaterial": (u"altrender", u"audience", u"id", u"type",),
    u"seriesstmt": (u"altrender", u"audience", u"encodinganalog", u"id",),
    u"sponsor": (u"altrender", u"audience", u"encodinganalog", u"id",),
    u"subject": (u"altrender", u"audience", u"authfilenumber",
                 u"encodinganalog", u"id", u"normal", u"rules", u"source",),
    u"subtitle": (u"altrender", u"audience", u"encodinganalog", u"id",),
    u"table": (u"altrender", u"audience", u"colspec", u"frame", u"id",
               u"pgwide", u"rowsep",),
    u"tbody": (u"altrender", u"audience", u"id", u"valign",),
    u"tgroup": (u"align", u"altrender", u"audience", u"colsep", u"id",
                u"rowsep",),
    u"thead": (u"altrender", u"audience", u"id", u"valign",),
    u"title": (u"actuate", u"altrender", u"arcrole", u"audience",
               u"authfilenumber", u"encodinganalog", u"entityref", u"href",
               u"id", u"linktype", u"normal", u"render", u"role", u"rules",
               u"show", u"source", u"title", u"type", u"xpointer",),
    u"titleproper": (u"altrender", u"audience", u"id", u"render",),
    u"titlestmt": (u"altrender", u"audience", u"encodinganalog", u"id",),
    u"unitdate": (u"altrender", u"audience", u"certainty", u"datechar", u"id",
                  u"label", u"type",),
    u"unitid": (u"altrender", u"audience", u"countrycode", u"id",
                u"identifier", u"label", u"repositorycode",),
    u"unittitle": (u"altrender", u"audience", u"id", u"label",),
    u"userrestrict": (u"altrender", u"audience", u"id",),
}


def _delete_attributes_and_log(xml_elt, attrs_to_delete, log_data):
    count_elts = 0
    count_atts = 0
    deleted = []
    for attname in xml_elt.attrib:
        if attname in attrs_to_delete:
            deleted.append(attname)
            xml_elt.attrib.pop(attname)
            count_elts = 1
            count_atts += 1
    if len(deleted) > 0:
        msg = u"    Deleted attributes: {0}".format(u" ".join(deleted))
        log_data.append(log_element(xml_elt, msg=msg))
    return count_elts, count_atts


def _delete_attributes(xml_elt, attrs_to_delete):
    count_elts = 0
    count_atts = 0
    for attname in xml_elt.attrib:
        if attname in attrs_to_delete:
            xml_elt.attrib.pop(attname)
            count_elts = 1
            count_atts += 1
    return count_elts, count_atts


class AttributesEraser(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"attributes-eraser"
    name = u"Erasing attributes that don't exist in Ape-EAD"
    category = u"Cleansing"
    desc = (
        u"Ape-EAD has more constraints than EAD-2002. Numerous elements "
        u"cannot contain some attributes that were authorized in EAD-2002. "
        u"This action deletes these attributes.")

    def _execute(self, xml_root, logger, log_details):
        count_elts, count_atts = 0, 0
        if log_details:
            log_data = []
            for elt in xml_root.iter(etree.Element):
                namespace, name = split_qname(elt.tag)
                if namespace is not None or len(elt.attrib) == 0:
                    continue
                cnt_e, cnt_a = _delete_attributes_and_log(
                    elt, ATTR_TO_ERASE.get(name, tuple()), log_data)
                count_elts += cnt_e
                count_atts += cnt_a
        else:
            for elt in xml_root.iter(etree.Element):
                namespace, name = split_qname(elt.tag)
                if namespace is not None or len(elt.attrib) == 0:
                    continue
                cnt_e, cnt_a = _delete_attributes(
                    elt, ATTR_TO_ERASE.get(name, tuple()))
                count_elts += cnt_e
                count_atts += cnt_a
        if count_elts > 0:
            logger.warning(
                u"Deleting non-legit attributes from the elements. {0:d} "
                u"attributes have been deleted from {1:d} elements."
                u"".format(count_elts, count_atts))
            if log_details:
                logger.warning(u"The following elements have had some "
                               u"attributes deleted:", u"\n".join(log_data))
        return xml_root


def _descgrp_children_move(descgrp_elt, first_following_sibling, log_data):
    count = 1
    if log_data is not None:
        log_data.append(log_element(descgrp_elt))
    last_inserted_note = None
    for child in descgrp_elt.iterchildren(etree.Element):
        nsp, name = split_qname(child.tag)
        if nsp is None and name == u"head":
            last_inserted_note = None
        elif nsp is None and name == u"descgrp":
            last_inserted_note = None
            count += _descgrp_children_move(child, first_following_sibling)
        elif nsp is None and name == u"note":
            last_inserted_note = None
            did_elt = descgrp_elt.find(u"did")
            if did_elt is None:
                continue
            did_elt.append(child)
        elif nsp is None and name in (u"address", u"blockquote", u"chronlist",
                                      u"list", u"p", u"table"):
            if last_inserted_note is None:
                did_elt = descgrp_elt.find(u"did")
                if did_elt is None:
                    continue
                last_inserted_note = etree.SubElement(did_elt, u"note",
                                                      type="descgrp")
            last_inserted_note.append(child)
        else:
            first_following_sibling.addprevious(child)
    return count


class DescgrpRemover(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"descgrp-remover"
    name = u"Removing the <descgrp> elements"
    category = u"Cleansing"
    desc = (u"The <descgrp> elements don't exist any more in Ape-EAD. "
            u"This action moves their children either into their parent "
            u"(<c> or <archdesc>) or into the <did> that exists inside this "
            u"parent.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        log_data = None
        if log_details:
            log_data = []
        for elt in xml_root.xpath(u'.//descgrp[not(parent::descgrp)]'):
            count += _descgrp_children_move(elt, elt, log_data)
            elt.getparent().remove(elt)
        if count > 0:
            logger.warning(
                u"Removing <descgrp> elements from the <archdesc> and <c> "
                u"by moving their child elements inside these <archdesc> or "
                u"<c> elements or inside the <did> that exists in these "
                u"<archdesc> or <c>. Please note that the <head> from the "
                u"<descgrp> have not been moved. {0:d} <descgrp> elements "
                u"have been emptied and deleted.".format(count))
            if log_details:
                logger.warning(u"The following elements have been removed:",
                               u"\n".join(log_data))
        return xml_root


ELTS_TO_ERASE = {u'heritedcontrolaccess', u'runner', u'index', u'xmlvalue'}


class ElementsEraser(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"elements-eraser"
    name = u"Erasing the unknown elements"
    category = u"Cleansing"
    desc = (u"Some elements don't exist in Ape-EAD. "
            u"This action deletes them and their content.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for nodename in ELTS_TO_ERASE:
            for node_elt in xml_root.xpath('.//' + nodename):
                if log_details:
                    log_data.append(log_element(node_elt, text_content=True))
                node_elt.getparent().remove(node_elt)
                count += 1
        if count > 0:
            logger.warning(
                u"{0:d} <runner> elements have been deleted.".format(count))
            if log_details:
                logger.warning(
                    u"The following elements have been deleted:",
                    u"\n".join(log_data))
        return xml_root



# -*- coding: utf-8 -*-
"""
Module defining the formats for Ape-EAD and EAD-2002 (instances of
:class:`~glamconv.transformer.formats.DataFormat)
"""

from glamconv.transformer.formats import DataFormat
from glamconv.transformer.rw_actions import XmlReader, XmlWriter
from glamconv.ead.validators import Ead2002Validator, ApeEadValidator


EAD_2002 = DataFormat(
    u"ead-2002", u"XML / EAD-2002", u"text/xml", u".xml",
    u"XML data complying with the EAD-2002 standard as defined in the EAD-2002"
    u" DTD")
XmlReader.applicable_for.add(EAD_2002)
XmlWriter.applicable_for.add(EAD_2002)
Ead2002Validator.applicable_for.add(EAD_2002)
EAD_2002.reading_class = XmlReader
EAD_2002.writing_class = XmlWriter
EAD_2002.validating_class = Ead2002Validator

EAD_APE = DataFormat(
    u"ape-ead", u"XML / Ape-EAD", u"text/xml", u".xml",
    u"XML data complying with the Ape-EAD standard as defined in the Ape-EAD "
    u"XML Schema")
XmlReader.applicable_for.add(EAD_APE)
XmlWriter.applicable_for.add(EAD_APE)
ApeEadValidator.applicable_for.add(EAD_APE)
EAD_APE.reading_class = XmlReader
EAD_APE.writing_class = XmlWriter
EAD_APE.validating_class = ApeEadValidator



# -*- coding: utf-8 -*-
"""
Module defining the processes and the steps.

A process is a sequence of steps, each step being an action with a
specification of parameter values for this action. A process can be be run,
it then executes each of the steps on the given input file.
"""

from traceback import format_exc
from glamconv.transformer.libraries import get_format, get_action
from glamconv.transformer.actions import WriteAction, ValidateAction
from glamconv.transformer.logger import RunLog


class Step(object):
    def __init__(self, action, log_details, specif_params):
        if specif_params is None:
            specif_params = {}
        self.action = action
        self.log_details = log_details
        self.params = specif_params


class Process(object):
    @classmethod
    def from_json(cls, data):
        proc = cls(data["in_dataformat"], data["out_dataformat"],
                   steps_specif=data["steps"])
        return proc

    def __init__(self, in_format_uid, out_format_uid, steps_specif=None):
        self.input_format = get_format(in_format_uid)
        if out_format_uid is not None:
            self.output_format = get_format(out_format_uid)
        else:
            self.output_format = None
        self.steps = []  # [Step(), ]
        if steps_specif is not None:
            self.build_steps(steps_specif)

    def build_steps(self, steps_specif):
        self.steps[:] = []  # Delete list content
        ending_idx = len(steps_specif)
        if self.output_format is not None:
            # Last two actions should be the validating action and the writing
            # action for the output format
            ending_idx = len(steps_specif)-2
        for idx, specif in enumerate(steps_specif):
            format_uid = self.input_format.uid
            if idx >= ending_idx:
                format_uid = self.output_format.uid
            try:
                action_class = get_action(format_uid, specif[u"id"])
            except KeyError:
                raise RuntimeError(
                    u"Can't find the action '{0}' for format '{1}'"
                    u"".format(specif[u"id"], format_uid))
            self.steps.append(
                Step(action_class(), specif[u"log_details"],
                     specif[u"params"])
            )

    def check(self):
        if len(self.steps) == 0:
            return True
        assert len(self.steps) > 2, \
            (u"Process should at least contain two steps for reading and "
             u"validating the file in input format")
        assert isinstance(self.steps[0].action,
                          self.input_format.reading_class), \
            (u"First step should read the file in input format")
        assert isinstance(self.steps[1].action,
                          self.input_format.validating_class), \
            (u"Second step should validate the file in input format")
        if self.output_format is not None:
            assert isinstance(self.steps[-1].action,
                              self.output_format.writing_class), \
                (u"Last step should write the file in output format")
            assert isinstance(self.steps[-2].action,
                              self.output_format.validating_class), \
                (u"Penultimate step should validate the file in output format")
        return True

    def run(self, input_stream, output_stream=None, progress_callback=None):
        logger = RunLog()
        logger.start_run()
        input_obj = input_stream
        output_obj = None
        for idx, step in enumerate(self.steps):
            if progress_callback is not None:
                progress_callback(idx)
            logger.start_step(idx)
            try:
                output_obj = step.action.run(input_obj, logger,
                                             step.log_details, step.params)
            except Exception:
                # An error has occured during the execution of the action.
                logger.error(
                    u"The following error has occured and has stopped "
                    u"the run:", format_exc())
                logger.failure = True
                logger.failure_step_index = idx
                logger.failure_message = (
                    u"The run was interrupted because of the following error "
                    u"during step #{0}:\n{1}".format(idx+1, format_exc()))
                logger.end_step()
                if idx > 1 and isinstance(self.steps[-1].action, WriteAction):
                    # We already have started the transformation. Tries to
                    # write the previous result (before the error) thanks to
                    # the last step (as it is a writing action).
                    logger.start_step(len(self.steps)-1)
                    logger.info(u"Trying to write the latest intermediate "
                                u"result...")
                    try:
                        last_step = self.steps[-1]
                        output_obj = last_step.action.run(input_obj, logger,
                                                          False,
                                                          last_step.params)
                        logger.info(u"The write operation succeed")
                    except Exception:
                        logger.error(
                            u"The write failed because of the following error:"
                            u"\n{0}".format(format_exc()))
                        output_obj = None
                    logger.end_step()
                # Stops the process.
                break
            # If the action is a validation action, takes into account the
            # result of this validation
            if isinstance(step.action, ValidateAction):
                # output_obj is the result of the validation
                if output_obj:
                    if idx == 1:
                        # The second step is the validation of the input file.
                        logger.input_validation = True
                    else:
                        logger.output_validation = True
                else:
                    _, msg, msg_data = logger.get_last_message()
                    if msg is not None:
                        if len(msg_data) > 256:
                            msg_data = (msg_data[:253]
                                        + u"...\nSee log for full message")
                    if idx == 1:
                        # The second step is the validation of the input file.
                        logger.input_validation = False
                        logger.input_validation_message = (
                            u"The input file is not valid. The following "
                            u"errors have been "
                            u"reported:\n{0}".format(msg_data))
                        logger.error(
                            u"The input validation has failed! "
                            u"Nevertheless, try to apply the steps...")
                    else:
                        logger.output_validation = False
                        logger.output_validation_message = (
                            u"The output file is not valid. The following "
                            u"errors have been "
                            u"reported:\n{0}".format(msg_data))
                        logger.error(u"The output validation has failed!")
            else:
                # For the actions other than the validation ones, the output
                # object becomes the input object of the next action.
                input_obj = output_obj
            logger.end_step()
        if output_stream is not None and output_obj is not None:
            output_stream.write(output_obj)
        logger.end_run()
        return logger


def build_validation_process(format_uid):
    prc = Process(format_uid, None)
    steps = [
        Step(prc.input_format.reading_class()),
        Step(prc.input_format.validating_class()),
    ]
    prc.steps.extend(steps)
    prc.check()
    return prc



# -*- coding: utf-8 -*-
"""
Script running the process for transforming EAD-2002 into Ape-EAD onto the
XML files found inside a hierarchy of directories.
"""

from os import path as osp
import argparse as argp
from glamconv.cli.commands import register_actions, transform_files_from_dir


HERE = osp.abspath(osp.dirname(__file__))

PROCESS_FILENAME = osp.join(HERE, "generic-ead-transformation.json")

COUNTRY_CODES = (
    u"AF AX AL DZ AS AD AO AI AQ AG AR AM AW AU AT AZ BS BH BD BB "
    u"BY BE BZ BJ BM BT BO BA BW BV BR IO BN BG BF BI KH CM CA CV "
    u"KY CF TD CL CN CX CC CO KM CG CD CK CR CI HR CU CY CZ DK DJ "
    u"DM DO EC EG SV GQ ER EE ET FK FO FJ FI FR GF PF TF GA GM GE "
    u"DE GH GI GR GL GD GP GU GT GN GW GY HT HM VA HN HK HU IS IN "
    u"ID IR IQ IE IL IT JM JP JO KZ KE KI KP KR KW KG LA LV LB LS "
    u"LR LY LI LT LU MO MK MG MW MY MV ML MT MH MQ MR MU YT MX FM "
    u"MD MC MN MS MA MZ MM NA NR NP NL AN NC NZ NI NE NG NU NF MP "
    u"NO OM PK PW PS PA PG PY PE PH PN PL PT PR QA RE RO RU RW SH "
    u"KN LC PM VC WS SM ST SA SN CS SC SL SG SK SI SB SO ZA GS ES "
    u"LK SD SR SJ SZ SE CH SY TW TJ TZ TH TL TG TK TO TT TN TR TM "
    u"TC TV UG UA AE GB US UM UY UZ VU VE VN VG VI WF EH YE ZM ZW "
    u"RS ME EU"
).split()

DEFAULT_COUNTRY = u"FR"


def id_params_setter(process, filename):
    identifiers_step = None
    for step in process.steps:
        if step.action.uid == u"identifiers-definer":
            identifiers_step = step
            break
    if identifiers_step is not None:
        # Setting agency code and country code
        code = osp.splitext(filename)[0].split("_")[0].upper()
        country = DEFAULT_COUNTRY
        if len(code) > 2:
            if code[:2] in COUNTRY_CODES:
                country = code[:2]
                code = code[2:]
        code = country + u"-" + code
        identifiers_step.params[u"agency_code"] = code
        identifiers_step.params[u"country_code"] = country
        identifiers_step.params[u"document_id"] = filename
        identifiers_step.params[u"favour_xml_document_id"] = True


def run(input_dir, output_dir):
    register_actions()
    input_dir = osp.normpath(osp.abspath(input_dir))
    output_dir = osp.normpath(osp.abspath(output_dir))
    transform_files_from_dir(
        PROCESS_FILENAME, input_dir, output_dir, recursive=True,
        adapt_process_params=id_params_setter)


if __name__ == "__main__":
    parser = argp.ArgumentParser(
        description=u"Transform the XML files in a directory tree from "
        u"EAD-2002 standard to Ape-EAD standard.")
    parser.add_argument(
        u"input_dir", metavar=u"input_directory", type=str,
        help=u"name of the directory where the XML files are "
        u"read from.")
    parser.add_argument(
        u"output_dir", metavar=u"output_directory", type=str,
        help=u"name of the directory where the transformed XML "
        u"files are written to. Should be different from the "
        u"input directory to avoid overwriting the input "
        u"files.")
    args = parser.parse_args()
    run(args.input_dir, args.output_dir)



# -*- coding: utf-8 -*-
"""
Module defining the classes used to log the run of a transformation process
in order to transform of a file into another file thanks to a series of steps
(action + parameter values).
"""

import datetime as dtm
import logging

from dateutil.tz import tzutc


ERROR = u"error"
"""
Type of log message describing an error

Type: :class:`str`
"""
WARNING = u"warning"
"""
Type of log message describing an error

Type: :class:`str`
"""
INFO = u"info"
"""
Type of log message describing an error

Type: :class:`str`
"""
EPOCH = dtm.datetime.fromtimestamp(0, tzutc())
"""
Datetime of the 1st January 1970 (Epoch).

Type: :class:`datetime.datetime`
"""


class RunLog(object):
    def __init__(self, stdlogger=None):
        self.start_timestamp = None
        self.steps_log = []  # StepLog
        self.end_timestamp = None
        self.input_validation = None
        self.input_validation_message = u""
        self.output_validation = None
        self.output_validation_message = u""
        self.failure = False
        self.failure_step_index = None
        self.failure_message = u""
        self._current_log = None
        self.stdlogger = stdlogger or logging.getLogger('ead.transform')

    @property
    def result(self):
        if self.failure:
            return True
        if self.output_validation is not None:
            return self.output_validation
        # output_validation can still be None if there was only two steps
        # for validating a file
        return self.input_validation or False

    def start_run(self):
        self.start_timestamp = dtm.datetime.now(tzutc())

    def end_run(self):
        self.end_timestamp = dtm.datetime.now(tzutc())

    def start_step(self, index):
        self._current_step_log = StepLog(index)
        self.steps_log.append(self._current_step_log)

    def end_step(self):
        self._current_step_log = None

    def action_start(self, name, param_values):
        if self._current_step_log is not None:
            self._current_step_log.set_action(name, param_values)
        else:
            raise RuntimeError(
                u"Can't log an action start as no step has begun")

    def info(self, message, data=u""):
        self.stdlogger.info(message + data)
        if self._current_step_log is not None:
            self._current_step_log.record_message(INFO, message, data)
        else:
            raise RuntimeError(u"Can't log a message as no step has begun")

    def warning(self, message, data=u""):
        self.stdlogger.warning(message + data)
        if self._current_step_log is not None:
            self._current_step_log.record_message(WARNING, message, data)
        else:
            raise RuntimeError(u"Can't log a message as no step has begun")

    def error(self, message, data=u""):
        self.stdlogger.error(message + data)
        if self._current_step_log is not None:
            self._current_step_log.record_message(ERROR, message, data)
        else:
            raise RuntimeError(u"Can't log a message as no step has begun")

    def get_last_message(self):
        if len(self.steps_log) > 0:
            last_step = self.steps_log[-1]
            if len(last_step.messages) > 0:
                return last_step.messages[-1]
        return None, None, None

    def dump(self):
        start_delta = self.start_timestamp - EPOCH
        start_msecs = int(start_delta.total_seconds() * 1000)
        end_delta = self.end_timestamp - EPOCH
        end_msecs = int(end_delta.total_seconds() * 1000)
        return {u"startTime": start_msecs,
                u"endTime": end_msecs,
                u"result": self.result,
                u"inputValidation": self.input_validation,
                u"inputValidationMessage": self.input_validation_message,
                u"outputValidation": self.output_validation,
                u"outputValidationMessage": self.output_validation_message,
                u"failure": self.failure,
                u"failureStepIndex": self.failure_step_index,
                u"failureMessage": self.failure_message,
                u"steps": [step.dump() for step in self.steps_log]}


class StepLog(object):
    def __init__(self, step_index):
        self.step_index = step_index
        self.action_name = u""
        self.action_params_val = []  # [ (param_name, param_val), ]
        self.messages = []  # [ (level, message_text, data_text), ]

    def set_action(self, name, param_values):
        self.action_name = name
        self.action_params_val = param_values

    def record_message(self, level, message, data):
        self.messages.append((level, message, data))

    def dump(self):
        data = {u"index": self.step_index,
                u"actionName": self.action_name,
                u"actionParamsVal": self.action_params_val,
                u"messages": []}
        for (level, msg, msg_data) in self.messages:
            data[u"messages"].append({u"level": level,
                                      u"message": msg,
                                      u"data": msg_data})
        return data



# -*- coding: utf-8 -*-
"""
Module defining basic read / write actions
"""

from lxml import etree
from six import binary_type

from glamconv.transformer.actions import ReadAction, WriteAction
from glamconv.ead.utils import NS


class XmlReader(ReadAction):
    uid = u"xml-reader"
    name = u"Reading of an XML file"
    desc = (u"This action reads an XML file and returns the XML data it "
            u"contains.")

    def _execute(self, input_flow, logger, log_details):
        if isinstance(input_flow, etree._Element):
            return input_flow
        if isinstance(input_flow, etree._ElementTree):
            return input_flow.getroot()
        tree = etree.parse(input_flow)
        xml_root = tree.getroot()
        return xml_root


class XmlWriter(WriteAction):
    uid = u"xml-writer"
    name = u"Writing an XML file"
    desc = (u"This action writes the XML data it receives into an XML file.")

    def _execute(self, xml_root, logger, log_details):
        # beautify namespaces in output: use ead ns as the default one
        nsmap = NS.copy()
        nsmap[None] = nsmap.pop('ead')
        new_root = etree.Element('ead', nsmap=nsmap)
        new_root[:] = xml_root[:]
        new_root.text = xml_root.text
        new_root.tail = xml_root.tail
        serialized = etree.tostring(
            new_root, encoding="UTF-8", xml_declaration=True, pretty_print=True
        )
        if isinstance(serialized, binary_type):
            # in Python3 return type of tostring is bytes;
            # encoding='unicode' could be used to serialize to str
            # but encoding='unicode' and xml_declaration=True
            # cannot be used at the same time
            return serialized.decode("utf-8")
        else:
            return serialized



# -*- coding: utf-8 -*-
"""
Module containing actions for converting, moving or correcting XML elements
that contain text data.
"""

import itertools
from lxml import etree
from glamconv.ead.utils import (log_element, write_hierarchy, split_qname,
                                insert_text_at_element_end,
                                insert_text_before_element,
                                adjust_content_in_mixed)
from glamconv.ead.formats import EAD_2002
from glamconv.transformer.actions import TransformAction


class AddressConverter(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"address-converter"
    name = u"Converting <address> into <p> in non-legit parents"
    category = u"Text data"
    desc = (
        u"In Ape-EAD, <address> elements can only occur in a few elements. "
        u"This action converts the <address> that occur outside these "
        u"legit parents into paragraphs (<p>). If the parent element can "
        u"only contain text data, this action directly adds the text in "
        u"this parent, each line separated by a hyphen. Be aware that in "
        u"case of imbricated elements, this action can produce a result "
        u"that have imbricated paragraphs (<p>); be sure to execute the "
        u"action that removes non-legit paragraphs after this action.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        text_only_parents = (u"entry", u"event", u"extref", u"extrefloc",
                             u"ref", u"refloc")
        parents = (u"accessrestrict", u"accruals", u"acqinfo",
                   u"altformavail", u"appraisal", u"arrangement",
                   u"bibliography", u"bioghist", u"blockquote",
                   u"controlaccess", u"custodhist",
                   u"descgrp", u"div", u"dsc", u"entry",
                   u"event", u"extref", u"extrefloc", u"fileplan",
                   u"item", u"note", u"odd", u"originalsloc",
                   u"otherfindaid", u"p", u"phystech", u"prefercite",
                   u"processinfo", u"ref", u"refloc", u"relatedmaterial",
                   u"scopecontent", u"separatedmaterial", u"userestrict")
        xpath_req = u' | '.join([u'.//{0}/address'.format(name)
                                 for name in parents])
        for addr in xml_root.xpath(xpath_req):
            parent = addr.getparent()
            if split_qname(parent.tag)[1] in text_only_parents:
                texts = []
                for adln in addr.iterchildren(etree.Element):
                    adjust_content_in_mixed(adln)
                    texts.append(adln.text)
                insert_text_before_element(addr, u" - ".join(texts))
                log_elt = parent
            else:
                p = etree.Element(u"p")
                addr.addprevious(p)
                for adln in addr.iterchildren(etree.Element):
                    insert_text_at_element_end(p, adln.text)
                    p.extend(adln.getchildren())
                    etree.SubElement(p, u"lb")
                if len(p) > 0:
                    p.remove(p[-1])  # Removes last <lb>
                log_elt = p
            parent.remove(addr)
            count += 1
            if log_details:
                log_data.append(log_element(log_elt, text_content=True))
        if count > 0:
            logger.warning(u"{0:d} <address> elements have been converted in "
                           u"<p> elements.".format(count))
            if log_details:
                logger.warning(u"The following elements have been converted:",
                               u"\n".join(log_data))
        return xml_root


def _gather_children_to_move(elt, elts_to_move):
    for child in elt.iterchildren(etree.Element):
        if child.tag == elt.tag:
            _gather_children_to_move(child, elts_to_move)
        else:
            elts_to_move.append(child)


class BlockquoteRemover(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"blockquote-remover"
    name = u"Removing <blockquote> elements"
    category = u"Text data"
    desc = (
        u"The <blockquote> elements don't exist anymore in Ape-EAD. "
        u"This action moves their children into the blockquote parent. The "
        u"blockquote is therefore removed from the EAD tree but its "
        u"content still exits in the tree. When possible, the texts inside "
        u"the paragraphs that were in the blockquote are encapsulated "
        u"into an <emph render=\"italic\"> element. Be aware that in case "
        u"of imbricated elements, this action can produce a result that "
        u"have imbricated paragraphs (<p>); be sure to execute the action "
        u"that removes non-legit paragraphs after this action.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for blq in xml_root.xpath(u'.//blockquote[not(parent::blockquote)]'):
            count += 1
            if log_details:
                log_data.append(log_element(blq, text_content=True))
            # Moves the subelements (children) into the blockquote parent
            elts_to_move = []
            _gather_children_to_move(blq, elts_to_move)
            for subelt in elts_to_move:
                # If subelt is an empty paragraph (most of the cases),
                # adds an emph in the paragraph for italizing the text
                if subelt.tag == u"p" and len(subelt) == 0:
                    emph = etree.Element(u'emph', render=u"italic")
                    emph.text = subelt.text
                    subelt.text = None
                    subelt.append(emph)
                blq.addprevious(subelt)
            # If blockquote has a tail text, inserts it into a paragraph
            if blq.tail is not None and len(blq.tail.strip()) > 0:
                p = etree.Element(u'p')
                p.text = blq.tail
                blq.tail = None
                blq.addprevious(p)
            # Deletes blockquote
            blq.getparent().remove(blq)
        if count > 0:
            logger.warning(
                u"{0:d} <blockquote> elements have been removed by moving "
                u"their child elements into the blockquote parent."
                u"".format(count))
            if log_details:
                logger.warning(u"The following elements have been emptied and "
                               u"deleted:", u"\n".join(log_data))
        return xml_root


class NoteRemover(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"note-remover"
    name = u"Removing <note> elements in non-legit parents"
    category = u"Text data"
    desc = (
        u"In Ape-EAD, <note> elements can only occur in a few elements. "
        u"This action moves the content of the <note> that occur outside "
        u"these legit elements into the note parent. The note is therefore "
        u"removed from the EAD tree but its content still exists in the "
        u"tree. Be aware that in case of imbricated elements, this action "
        u"can produce a result that have imbricated paragraphs (<p>); be "
        u"sure to execute the action that removes non-legit paragraphs "
        u"after this action.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        parents = (u"accessrestrict", u"accruals", u"acqinfo",
                   u"altformavail", u"appraisal",
                   u"archref", u"arrangement", u"bibliography",
                   u"bioghist", u"blockquote", u"controlaccess",
                   u"custodhist", u"descgrp", u"div", u"dsc",
                   u"entry", u"event", u"extref", u"extrefloc",
                   u"fileplan", u"item", u"odd", u"originalsloc",
                   u"otherfindaid", u"phystech", u"prefercite",
                   u"processinfo", u"ref", u"refloc",
                   u"relatedmaterial", u"scopecontent",
                   u"separatedmaterial", u"userestrict")
        xpath_req = u' | '.join([u'.//{0}/note'.format(name)
                                 for name in parents])
        for note in xml_root.xpath(xpath_req):
            count += 1
            if log_details:
                log_data.append(log_element(note, text_content=True))
            # Moves the subelements (children) into the note parent
            elts_to_move = []
            _gather_children_to_move(note, elts_to_move)
            for subelt in elts_to_move:
                note.addprevious(subelt)
            # If note has a tail text, inserts it into a paragraph
            if note.tail is not None and len(note.tail.strip()) > 0:
                p = etree.Element(u'p')
                p.text = note.tail
                note.tail = None
                note.addprevious(p)
            # Deletes note
            note.getparent().remove(note)
        if count > 0:
            logger.warning(
                u"{0:d} <note> elements have been removed by moving their "
                u"child elements into the note parent.".format(count))
            if log_details:
                logger.warning(u"The following elements have been emptied "
                               u"and deleted:", u"\n".join(log_data))
        return xml_root


class ChronlistConverter(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"chronlist-converter"
    name = u"Converting <chronlist> into regular lists"
    category = u"Text data"
    desc = (u"In Ape-EAD, the <chronlist> elements don't exist anymore. "
            u"This action converts these chronological lists into regular "
            u"lists, each item containing the date and the associated "
            u"events. If possible, the date is inserted in an "
            u"<emph render=\"bold\"> element.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for list_elt in xml_root.xpath(u'.//chronlist'):
            list_elt.tag = u"list"
            # Header
            lsthead = list_elt.find(u"listhead")
            if lsthead is not None:
                new_head = False
                head = list_elt.find(u"head")
                if head is None:
                    new_head = True
                    head = etree.Element(u"head")
                for hd_elt in lsthead.iterchildren(etree.Element):
                    if head.text or len(head) > 0:
                        insert_text_at_element_end(head, u" / ")
                    insert_text_at_element_end(head, hd_elt.text)
                    head.extend(hd_elt.getchildren())
                if new_head and (head.text or len(head) > 0):
                    list_elt.insert(0, head)
                list_elt.remove(lsthead)
            # Items
            for chritm in list_elt.xpath(u'chronitem'):
                item = etree.Element(u"item")
                chritm.addprevious(item)
                date = chritm.find(u"date")
                if date is None:
                    continue
                if len(date) == 0:
                    emph = etree.Element(u"emph", render=u"bold")
                    emph.text = date.text
                    date.text = None
                    date.append(emph)
                insert_text_at_element_end(item, date.text)
                item.extend(date.getchildren())
                for idx, evt in enumerate(
                        chritm.xpath(u'event|eventgrp/event')):
                    insert_text_at_element_end(item,
                                               u" : " if idx == 0 else u" ; ")
                    insert_text_at_element_end(item, evt.text)
                    item.extend(evt.getchildren())
                list_elt.remove(chritm)
            count += 1
            if log_details:
                log_data.append(log_element(list_elt))
        if count > 0:
            logger.warning(u"{0:d} <chronlist> elements have been converted "
                           u"into regular lists containing <item>."
                           u"".format(count))
            if log_details:
                logger.warning(u"The following elements have been converted:",
                               u"\n".join(log_data))
        return xml_root


class DefinitionListConverter(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"def-list-converter"
    name = u"Converting <list> with <defitem> children into regular lists"
    category = u"Text data"
    desc = (u"In Ape-EAD, <list> elements can not have <defitem> children. "
            u"This action converts these lists into regular lists by "
            u"converting the <defitem> children into <item> children. "
            u"The content of the <label> child of <defitem> is inserted "
            u"at the begining of the <item> child of <defitem>. If possible, "
            u"the label is inserted in an <emph render=\"bold\"> element.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for list_elt in xml_root.xpath(u'.//list[defitem]'):
            # Header
            head = etree.Element(u"head")
            for hd_elt in list_elt.xpath(u'listhead/*'):
                if head.text or len(head) > 0:
                    insert_text_at_element_end(head, u" / ")
                insert_text_at_element_end(head, hd_elt.text)
                head.extend(hd_elt.getchildren())
            if head.text or len(head) > 0:
                list_elt.insert(0, head)
            lsth = list_elt.find(u"listhead")
            if lsth is not None:
                list_elt.remove(lsth)
            # Items
            for defitm in list_elt.xpath(u'defitem'):
                label = defitm.find(u"label")
                item = defitm.find(u"item")
                if item is None:
                    item = etree.Element(u"item")
                    defitm.append(item)
                if label is not None and len(label) == 0:
                    emph = etree.Element(u"emph", render=u"bold")
                    emph.text = label.text
                    label.text = None
                    label.append(emph)
                insert_text_at_element_end(label,
                                           (u" : " + (item.text or u"")))
                for child in reversed(list(label.getchildren())):
                    item.insert(0, child)
                item.text = label.text
                defitm.addprevious(item)
                list_elt.remove(defitm)
            count += 1
            if log_details:
                log_data.append(log_element(list_elt))
        if count > 0:
            logger.warning(u"{0:d} <list> elements containing <defitem> "
                           u"children have been converted into regular lists "
                           u"containing <item>.".format(count))
            if log_details:
                logger.warning(u"The following elements have been converted:",
                               u"\n".join(log_data))
        return xml_root


def _list_content_extractor(list_elt, extracted_elts, prefix=u"- ",
                            indent=u"  ", level=0):
    for child in list_elt.iterchildren(etree.Element):
        if child.tag == u"head":
            adjust_content_in_mixed(child)
            emph = etree.Element(u"emph", render=u"bold")
            emph.text = child.text
            child.text = indent*level
            child.append(emph)
            extracted_elts.append(child)
            continue
        if child.text or (len(child) > 0 and child[0].tag != u"list"):
            child.text = (indent*level) + prefix + (child.text or u"")
        sublist = child.find(u"list")
        while sublist is not None:
            itm = etree.Element(child.tag)
            itm.text = child.text
            child.text = None
            for subelt in reversed(list(sublist.itersiblings(preceding=True))):
                itm.append(subelt)
            if itm.text or len(itm) > 0:
                extracted_elts.append(itm)
            _list_content_extractor(sublist, extracted_elts,
                                    prefix, indent, level+1)
            child.remove(sublist)
            child.text = sublist.tail
            if child.text or(len(child) > 0 and child[0].tag != u"list"):
                child.text = (indent*level) + prefix + (child.text or u"")
            sublist = child.find(u"list")
        if child.text or len(child) > 0:
            extracted_elts.append(child)


class ListConverter(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"list-converter"
    name = u"Removing <list> elements in non-legit parents"
    category = u"Text data"
    desc = (u"In Ape-EAD, <list> elements can not occur anymore in some "
            u"elements. This action converts each of these lists into a "
            u"paragraph (<p>); the items being separated by line breaks "
            u"(<lb>). If the parent element can't contain a paragraph, "
            u"this action directly adds the text in this parent and "
            u"separates each item by a semi-colon. Be aware that in case "
            u"of imbricated elements, this action can produce a result that "
            u"have imbricated paragraphs (<p>); be sure to execute the action "
            u"that removes non-legit paragraphs after this action.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        text_only_parents = (u"event", u"extref", u"extrefloc", u"ref",
                             u"refloc")
        parents = (u"controlaccess", u"descgrp", u"entry", u"event",
                   u"extref", u"extrefloc", u"note", u"p",
                   u"phystech", u"ref", u"refloc")
        xpath_req = u' | '.join([u'.//{0}/list'.format(name)
                                 for name in parents])
        for list_elt in xml_root.xpath(xpath_req):
            parent = list_elt.getparent()
            if split_qname(parent.tag)[1] in text_only_parents:
                texts = []
                extracted_elts = []
                _list_content_extractor(list_elt, extracted_elts, prefix=u"",
                                        indent=u"")
                for elt in extracted_elts:
                    adjust_content_in_mixed(elt)
                    texts.append(elt.text)
                insert_text_before_element(list_elt, u" ; ".join(texts))
                log_elt = parent
            else:
                p = etree.Element(u"p")
                list_elt.addprevious(p)
                extracted_elts = []
                _list_content_extractor(list_elt, extracted_elts)
                for idx, elt in enumerate(extracted_elts):
                    if idx != 0:
                        etree.SubElement(p, u"lb")
                    insert_text_at_element_end(p, elt.text)
                    p.extend(elt.getchildren())
                log_elt = p
            parent.remove(list_elt)
            count += 1
            if log_details:
                log_data.append(log_element(log_elt, text_content=True))
        if count > 0:
            logger.warning(u"{0:d} <list> elements have been converted into "
                           u"<p> elements.".format(count))
            if log_details:
                logger.warning(u"The following elements have been converted:",
                               u"\n".join(log_data))
        return xml_root


class LegalstatusConverter(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"legalstatus-converter"
    name = u"Converting the <legalstatus> into <p>"
    category = u"Text data"
    desc = (u"The <legalstatus> elements don't exist in Ape-EAD. This action "
            u"converts these elements into paragraphs (<p>).")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for stat in xml_root.xpath(u'.//legalstatus'):
            stat.tag = u'p'
            count += 1
            if log_details:
                log_data.append(log_element(stat, text_content=True))
        if count > 0:
            logger.warning(u"{0:d} <legalstatus> elements have been converted "
                           u"into <p> elements.".format(count))
            if log_details:
                logger.warning(u"The following elements have been converted:",
                               u"\n".join(log_data))
        return xml_root


class ParagraphRemover(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"paragraph-remover"
    name = u"Removing <p> elements in non-legit parents"
    category = u"Text data"
    desc = (
        u"After executing the actions that removes blockquotes, addresses, "
        u"notes, lists or tables, the resulting tree can have paragraphs "
        u"that occur in non-legit parents (typically inside another "
        u"paragraph). This action removes these paragraphs but keeps their "
        u"content and adds a line break (<lb>) at the end of this content.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        for p_elt in xml_root.xpath(u'p[parent::p]'):
            insert_text_before_element(p_elt, p_elt.text)
            for child in p_elt:
                p_elt.addprevious(child)
            p_elt.addprevious(etree.Element(u"lb"))
            count += 1
        if count > 0:
            logger.warning(
                u"Removing <p> occuring inside non-legit parents: {0} "
                u"paragraphs have been removed.".format(count))
        return xml_root


MIXED_CONTENT = {
    u"abstract": tuple(),
    u"addressline": tuple(),
    u"archref": tuple(),
    u"author": tuple(),
    u"bibref": (u"imprint", u"name", u"title",),
    u"bibseries": tuple(),
    u"container": tuple(),
    u"corpname": tuple(),
    u"creation": (u"date",),
    u"date": tuple(),
    u"descrules": (u"extref",),
    u"dimensions": tuple(),
    u"edition": tuple(),
    u"emph": tuple(),
    u"entry": tuple(),
    u"event": tuple(),
    u"extent": tuple(),
    u"extref": tuple(),
    u"extrefloc": tuple(),
    u"famname": tuple(),
    u"function": tuple(),
    u"genreform": tuple(),
    u"geogname": tuple(),
    u"head": tuple(),
    u"head01": tuple(),
    u"head02": tuple(),
    u"imprint": (u"date", u"geogname", u"publisher",),
    u"item": (u"emph", u"extref", u"lb", u"list",),
    u"label": tuple(),
    u"langmaterial": (u"language",),
    u"language": tuple(),
    u"langusage": (u"language",),
    u"legalstatus": tuple(),
    u"materialspec": tuple(),
    u"name": tuple(),
    u"num": tuple(),
    u"occupation": tuple(),
    u"origination": (u"corpname", u"famname", u"name", u"persname",),
    u"p": (u"abbr", u"emph", u"expan", u"extref", u"lb", u"note",),
    u"persname": tuple(),
    u"physdesc": (u"dimensions", u"extent", u"genreform", u"physfacet",),
    u"physfacet": tuple(),
    u"physloc": tuple(),
    u"publisher": tuple(),
    u"ref": tuple(),
    u"refloc": tuple(),
    u"repository": (u"address", u"corpname", u"extref", u"name",),
    u"resource": tuple(),
    u"runner": tuple(),
    u"sponsor": tuple(),
    u"subarea": tuple(),
    u"subject": tuple(),
    u"subtitle": (u"emph", u"lb",),
    u"title": tuple(),
    u"titleproper": (u"emph", u"lb",),
    u"unitdate": tuple(),
    u"unitid": (u"extptr", u"title", u"abbr", u"emph", u"expan", u"lb",),
    u"unittitle": (u"abbr", u"emph", u"expan", u"lb",),
}


def _adjust_mixed_content(xml_elt):
    """
    Walk depth-first in the XML tree from ``xml_elt`` and adjust the mixed
    content of the XML element (see ``MIXED_CONTENT`` dictionary to know
    which element can be kept at a given level of the tree).
    """
    count = 0
    child_names = set()
    if len(xml_elt) == 0:
        return 0
    for child in xml_elt.iterchildren(etree.Element):
        child_names.add(split_qname(child.tag)[1])
        # First process the children (process depth-first in the tree)
        count += _adjust_mixed_content(child)
    namespace, name = split_qname(xml_elt.tag)
    to_keep = MIXED_CONTENT.get(name)
    if namespace is not None or to_keep is None:
        # ``xml_elt`` is not in ``MIXED_CONTENT`` and thus doesn't need to be
        # adjusted
        return count
    # If <lb> is not to keep, adds a space
    if u"lb" not in to_keep:
        for lb_elt in xml_elt.xpath(u'lb'):
            lb_elt.tail = u" " + (lb_elt.tail or u"")
    # Suppress the mixed elements not to be kept
    if len(child_names.difference(set(to_keep))) > 0:
        count += 1
    adjust_content_in_mixed(xml_elt, to_keep)
    return count


def _adjust_mixed_content_and_log(xml_elt):
    if len(xml_elt) == 0:
        return []
    children_log_data = {}
    child_names = set()
    for child in xml_elt.iterchildren(etree.Element):
        _, cname = split_qname(child.tag)
        child_names.add(cname)
        if cname not in children_log_data:
            children_log_data[cname] = []
        children_log_data[cname].extend(_adjust_mixed_content_and_log(child))
    namespace, name = split_qname(xml_elt.tag)
    to_keep = MIXED_CONTENT.get(name)
    if namespace is not None or to_keep is None:
        log = list(itertools.chain.from_iterable(children_log_data.values()))
        return log
    removed = set()
    # If <lb> is not to keep, adds a space
    if u"lb" not in to_keep:
        for lb_elt in xml_elt.xpath(u'lb'):
            lb_elt.tail = u" " + (lb_elt.tail or u"")
    # Suppress the mixed elements not to be kept
    for cname in child_names.difference(set(to_keep)):
        removed.add(cname)
        for item in children_log_data.pop(cname):
            removed = removed.union(item[1])
    adjust_content_in_mixed(xml_elt, to_keep)
    log_data = [(write_hierarchy(xml_elt), removed), ]
    log_data.extend(
        list(itertools.chain.from_iterable(children_log_data.values()))
    )
    return log_data


class MixedContentAdjuster(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"mixed-content-adjuster"
    name = u"Removing the non-legit children in the mixed content elements"
    category = u"Text data"
    desc = (u"In Ape-EAD, the elements with a mixed content (child elements "
            u"mixed with text content) have a much more restricted list of "
            u"allowed children. This action removes the non-legit children "
            u"but keeps their textual content.")

    def _execute(self, xml_root, logger, log_details):
        if log_details:
            log_data = _adjust_mixed_content_and_log(xml_root)
            count = len(log_data)
        else:
            count = _adjust_mixed_content(xml_root)
        if count > 0:
            logger.warning(
                u"{0:d} mixed content elements have had some of their "
                u"children deleted".format(count))
            if log_details:
                log_info = [u"{0}\n   Removed elements: {1}"
                            u"".format(elt, u", ".join(remov))
                            for elt, remov in log_data if len(remov) > 0]
                logger.warning(
                    u"The following mixed content elements have "
                    u"been modified:", u"\n".join(log_info))
        return xml_root



# -*- coding: utf-8 -*-
"""
Module containing actions for correcting XML elements that don't meet the
constraints described in EAD 2002 format.
"""

import re

from lxml import etree

from glamconv.ead.utils import log_element, split_qname
from glamconv.ead.formats import EAD_2002
from glamconv.transformer.actions import TransformAction


class ChangeChildrenOrderer(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"change-children-orderer"
    name = u"Ordering <change> children"
    category = u"EAD 2002 correction"
    desc = (u"In EAD, the children of <change> must occur in a given order "
            u"(first the <date>, then the <item> elements).")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        xpath_req = u".//change[date/preceding-sibling::*]"
        for change in xml_root.xpath(xpath_req):
            if log_details:
                log_data.append(log_element(change))
            count += 1
            # Puts the only <date> at beginning
            for date_elt in change.xpath(u"date[preceding-sibling::*]"):
                change.insert(0, date_elt)
        if count > 0:
            logger.info(
                u"Re-ordering children in <change> elements. {0:d} "
                u"elements have been corrected.".format(count))
            if log_details:
                logger.info(u"The following elements have been corrected:",
                            u"\n".join(log_data))
        return xml_root


class EadheaderChildrenOrderer(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"eadheader-children-orderer"
    name = u"Ordering <eadheader> children"
    category = u"EAD 2002 correction"
    desc = (u"In EAD, the children of <eadheader> must occur in a given order "
            u"(first the <eadid>, then <filedesc>, then <profiledesc>?, "
            u"then <revisiondesc>?).")

    def _execute(self, xml_root, logger, log_details):
        eadheader = xml_root.find('eadheader')
        if eadheader is None:
            return xml_root
        for nodename in ('revisiondesc', 'profiledesc', 'filedesc', 'eadid'):
            node = eadheader.find(nodename)
            if node is not None:
                eadheader.insert(0, node)
        return xml_root


class ArchdescChildrenOrderer(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"archdesc-children-orderer"
    name = u"Ordering <archdesc> children"
    category = u"EAD 2002 correction"
    desc = (
        u"In EAD, the children of <archdesc> must occur in a given order "
        u"(first the <did>, then the archive descriptive elements and then "
        u"the <dsc>). This action puts the children in the expected order. "
        u"This action can be applied to correct EAD trees that are not "
        u"valid against the EAD-2002 standard.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        xpath_req = (u".//archdesc[ "
                     u"  did/preceding-sibling::* or "
                     u"  dsc/following-sibling::*[not(self::dsc)]"
                     u"]")
        for elt in xml_root.xpath(xpath_req):
            if log_details:
                log_data.append(log_element(elt))
            count += 1
            # Puts the only <did> at beginning
            for did in elt.xpath(u"did[preceding-sibling::*]"):
                elt.insert(0, did)
            # Puts last elements before all <dsc>
            dsc = elt.find(u'dsc')
            for subelt in elt.xpath(
                    u"*[not(self::dsc) and preceding-sibling::dsc]"):
                dsc.addprevious(subelt)
        if count > 0:
            logger.info(
                u"Re-ordering children in <archdesc> elements. {0:d} "
                u"elements have been corrected.".format(count))
            if log_details:
                logger.info(u"The following elements have been corrected:",
                            u"\n".join(log_data))
        return xml_root


class CChildrenOrderer(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"c-children-orderer"
    name = u"Ordering <c> children"
    category = u"EAD 2002 correction"
    desc = (u"In EAD, the children of <c> must occur in a given order (first "
            u"the <did>, then the elements for archive description and then "
            u"the <c>). This action puts the children in the expected order."
            u"This action can be applied to correct EAD trees that are not "
            u"valid against the EAD-2002 standard.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        xpath_req = (u".//c["
                     u"  did/preceding-sibling::* or "
                     u"  c/following-sibling::*[not(self::c)]"
                     u"]")
        for elt in xml_root.xpath(xpath_req):
            count += 1
            if log_details:
                log_data.append(log_element(elt))
            # Puts the only <did> at beginning
            for did in elt.xpath(u"did[preceding-sibling::*]"):
                elt.insert(0, did)
            # Puts last elements before all <c>
            c = elt.find(u'c')
            for subelt in elt.xpath(
                    u"*[not(self::c) and preceding-sibling::c]"):
                c.addprevious(subelt)
        if count > 0:
            logger.info(
                u"Re-ordering children in <c> elements. {0:d} "
                u"elements have been corrected.".format(count))
            if log_details:
                logger.info(u"The following elements have been corrected:",
                            u"\n".join(log_data))
        return xml_root


EMPTY_ELTS_TO_ERASE = (
    u"accessrestrict", u"accruals", u"acqinfo", u"address", u"altformavail",
    u"appraisal", u"arrangement", u"bibliography", u"bioghist", u"c",
    u"change", u"controlaccess", u"custodhist", u"descgrp", u"did", u"dsc",
    u"fileplan", u"list", u"note", u"notestmt", u"odd", u"originalsloc",
    u"otherfindaid", u"profiledesc", u"publicationstmt", u"phystech",
    u"prefercite", u"processinfo", u"relatedmaterial", u"revisiondesc",
    u"scopecontent", u"separatedmaterial", u"seriesstmt", u"table",
    u"userestrict",)


class EmptyEltsEraser(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"empty-elts-eraser"
    name = u"Erasing empty elements"
    category = u"EAD 2002 correction"
    desc = (u"In EAD, numerous elements must have at least one child. This "
            u"action erases those elements that don't have any child element. "
            u"This action can be applied to correct EAD trees that are not "
            u"valid against the EAD-2002 format.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        xpath_req = u" | ".join([u".//{0}[not(*)]".format(eltname)
                                 for eltname in EMPTY_ELTS_TO_ERASE])
        for elt in xml_root.xpath(xpath_req):
            count += 1
            if log_details:
                log_data.append(log_element(elt))
            elt.getparent().remove(elt)
        if count > 0:
            logger.warning(u"{0:d} empty elements have been deleted."
                           u"".format(count))
            if log_details:
                logger.warning(u"The following empty elements have been"
                               u"deleted:", u"\n".join(log_data))
        return xml_root


ONE_CONTENT_CHILD_ELTS = (
    u"accessrestrict", u"accruals", u"acqinfo", u"altformavail",
    u"appraisal", u"arrangement", u"bibliography", u"bioghist",
    u"controlaccess", u"custodhist", u"descgrp", u"dsc",
    u"fileplan", u"odd", u"originalsloc", u"otherfindaid", u"phystech",
    u"prefercite", u"processinfo", u"relatedmaterial", u"scopecontent",
    u"separatedmaterial", u"seriesstmt", u"userestrict",)


class ParagraphAdder(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"paragraph-adder"
    name = u"Adding paragraph in elements with only a <head>"
    category = u"EAD 2002 correction"
    desc = (u"In EAD, numerous elements must have at least one child element "
            u"other than <head>. This action adds an empty paragraph (<p>) "
            u"in the elements that don't have such a child element. "
            u"This action can be applied to correct EAD trees that are not "
            u"valid against the EAD-2002 standard.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        xpath_req = u" | ".join(
            [u".//{0}[not(*[not(self::head)])]".format(name)
             for name in ONE_CONTENT_CHILD_ELTS])
        for elt in xml_root.xpath(xpath_req):
            count += 1
            if log_details:
                log_data.append(log_element(elt, text_content=True))
            etree.SubElement(elt, u"p")
        if count > 0:
            logger.warning(u"In {0:d} elements, an empty paragraph has been "
                           u"added in order to have at least one \"content\" "
                           u"child.".format(count))
            if log_details:
                logger.warning(u"The following elements have been modified:",
                               u"\n".join(log_data))
        return xml_root


EMPTY_ATTRS_TO_ERASE = {
    u"c": (u"level",),
    u"c01": (u"level",),
    u"c02": (u"level",),
    u"c03": (u"level",),
    u"c04": (u"level",),
    u"c05": (u"level",),
    u"c06": (u"level",),
    u"c07": (u"level",),
    u"c08": (u"level",),
    u"c09": (u"level",),
    u"c10": (u"level",),
    u"c11": (u"level",),
    u"c12": (u"level",),
}


class EmptyAttrsEraser(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"empty-attrs-eraser"
    name = u"Erasing empty attributes"
    category = u"EAD 2002 correction"
    desc = (u"In EAD, some attributes must have a value taken from a "
            u"vocabulary. This action erases these attributes that are empty."
            u"This action can be applied to correct EAD trees that are not "
            u"valid against the EAD-2002 standard.")

    def _execute(self, xml_root, logger, log_details):
        count_atts = 0
        count_elts = 0
        if log_details:
            log_data = []
        xpath_reqs = []
        for eltname, attnames in EMPTY_ATTRS_TO_ERASE.items():
            xpath_pred = u" or ".join([u'@{0}[normalize-space(text())=""]'
                                       u''.format(name) for name in attnames])
            xpath_reqs.append(u'.//{0}[{1}]'.format(eltname, xpath_pred))
        for elt in xml_root.xpath(u"|".join(xpath_reqs)):
            count_elts += 1
            if log_details:
                deleted = []
            for attname in EMPTY_ATTRS_TO_ERASE[split_qname(elt.tag)[1]]:
                value = elt.attrib.pop(attname, None)
                if value is not None:
                    count_atts += 1
                    if log_details:
                        deleted.append(attname)
            if log_details:
                msg = u"    Deleted attributes: {0}".format(u" ".join(deleted))
                log_data.append(log_element(elt, msg=msg))
        if count_elts > 0:
            logger.warning(u"{0:d} empty attributes have been deleted from "
                           u"{1:d} elements.".format(count_atts, count_elts))
            if log_details:
                logger.warning(u"The following elements have had some "
                               u"attributes deleted:", u"\n".join(log_data))
        return xml_root


class IdentifierConverter(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"identifier-converter"
    name = u"Converting identifier attributes"
    category = u"EAD 2002 correction"
    desc = (u"In EAD, except in <eadid> and <unitid>, the identifier must be "
            u"defined in an 'id' attribute. In <eadid> and <unitid>, it must "
            u"be defined in an 'identifier' attribute. Depending on the "
            u"element, this action renames the 'identifier' attribute in 'id' "
            u"(if there is not already an 'id' attribute) and the 'id' "
            u"attribute in 'identifier' (if there is not already an "
            u"'identifier' attribute). This action can be applied to correct "
            u"EAD trees that are not valid against the EAD-2002 standard.")

    def _execute(self, xml_root, logger, log_details):
        # identifier to id
        count1 = 0
        count2 = 0
        if log_details:
            log_data1 = []
            log_data2 = []
        for elt in xml_root.xpath(u'.//*[namespace-uri() = "" '
                                  u'     and not(self::unitid|self::eadid) '
                                  u'     and @identifier]'):
            if u"id" not in elt.attrib:
                elt.set(u"id", elt.attrib.pop(u"identifier"))
                count1 += 1
                if log_details:
                    log_data1.append(log_element(elt, attributes=(u"id",)))
            else:
                count2 += 1
                if log_details:
                    log_data2.append(
                        log_element(elt, attributes=(u"id", u"identifier",)))
                elt.attrib.pop(u"identifier")
        if count1 > 0:
            logger.warning(u"{0:d} elements have had an 'id' attribute "
                           u"defined from their 'identifier' attribute."
                           u"".format(count1))
            if log_details:
                logger.warning(u"The following elements have been corrected:",
                               u"\n".join(log_data1))
        if count2 > 0:
            logger.warning(u"{0:d} elements have had their 'identifier' "
                           u"attribute deleted because they already had an "
                           u"'id' attribute.".format(count2))
            if log_details:
                logger.warning(u"The following elements have been corrected:",
                               u"\n".join(log_data2))
        # id to identifier
        count1 = 0
        count2 = 0
        if log_details:
            log_data1 = []
            log_data2 = []
        for elt in xml_root.xpath(
                u".//*[(self::unitid|self::eadid) and @id]"):
            if u"identifier" not in elt.attrib:
                elt.set(u"identifier", elt.attrib.pop(u"id"))
                count1 += 1
                if log_details:
                    log_data1.append(
                        log_element(elt, attributes=(u"identifier",)))
            else:
                count2 += 1
                if log_details:
                    log_data2.append(
                        log_element(elt, attributes=(u"id", u"identifier",)))
                elt.attrib.pop(u"id")
        if count1 > 0:
            logger.warning(
                u"{0:d} elements have had an 'identifier' attribute "
                u"defined from their 'id' attribute."
                u"".format(count1))
            if log_details:
                logger.warning(u"The following elements have been corrected:",
                               u"\n".join(log_data1))
        if count2 > 0:
            logger.warning(u"{0:d} elements have had their 'id' attribute "
                           u"deleted because they already had an 'identifier' "
                           u"attribute.".format(count2))
            if log_details:
                logger.warning(u"The following elements have been corrected:",
                               u"\n".join(log_data2))
        return xml_root


class UnitdateCertainlyConverter(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"unitdate-certainly-converter"
    name = u"Converting certainly attribute of <unitdate>"
    category = u"EAD 2002 correction"
    desc = (
        u"In EAD, the attribute for describing the certainty of a "
        u"<unitdate> is 'certainty'. This action corrects the name of this "
        u"attribute when it is mispelled 'certainly'. This action can be "
        u"applied to correct EAD trees that are not valid against the "
        u"EAD-2002 standard.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for elt in xml_root.xpath(u".//unitdate[@certainly]"):
            elt.set(u"certainty", elt.attrib.pop(u"certainly"))
            count += 1
            if log_details:
                log_data.append(log_element(elt, attributes=(u"certainty",),
                                            text_content=True))
        if count > 0:
            logger.warning(u"{0:d} elements have had a 'certainty' attribute "
                           u"defined from their 'certainly' attribute."
                           u"".format(count))
            if log_details:
                logger.warning(u"The following elements have been corrected:",
                               u"\n".join(log_data))
        return xml_root


class COTHERlevelConverter(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"c-OTHERlevel-converter"
    name = u"Converting OTHERlevel attribute of <c>"
    category = u"EAD 2002 correction"
    desc = (u"In EAD, the attribute for describing the other level inside a "
            u"<c> is 'otherlevel'. This action corrects the name of this "
            u"attribute when it is mispelled 'OTHERlevel'. This action can be "
            u"applied to correct EAD trees that are not valid against the "
            u"EAD-2002 standard.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for elt in xml_root.xpath(u".//c[@OTHERlevel]"):
            elt.set(u"otherlevel", elt.attrib.pop(u"OTHERlevel"))
            count += 1
            if log_details:
                log_data.append(log_element(elt, attributes=(u"otherlevel",)))
        if count > 0:
            logger.warning(
                u"{0:d} elements have had an 'otherlevel' attribute "
                u"defined from their 'OTHERlevel' attribute."
                u"".format(count))
            if log_details:
                logger.warning(u"The following elements have been corrected:",
                               u"\n".join(log_data))
        return xml_root


class LowercaseElements(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"lowercase-elts"
    name = u"lowercase elements"
    category = u"EAD 2002 correction"
    desc = u"In EAD, all elements should be lowercase."

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for node in xml_root.getiterator():
            if node.tag is etree.Comment:
                continue
            if not node.tag.islower():
                count += 1
                if log_details:
                    log_data.append(log_element(node))
            node.tag = node.tag.lower()
        if count > 0:
            logger.warning(u"{0:d} elements got lowercased", u"".format(count))
            if log_details:
                logger.warning(u"The following elements have been corrected:",
                               u"\n".join(log_data))
        return xml_root


VALID_ID_RGX = re.compile('[_a-z][_a-z0-9.:-]*', re.I | re.U)


def normalize_xml_id(xml_id):
    """try to build a valid xml identifier out of ``xml_id``"""
    chunks = (x.strip() for x in xml_id.split())
    chunks = (x for x in chunks if x != u'-')
    xml_id = u'-'.join(chunks)
    if xml_id and xml_id[0].isdigit():
        xml_id = '_' + xml_id
    return xml_id


class NormalizeIdAttributes(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"normalize-id"
    name = u"normalize id attributes"
    category = u"EAD 2002 correction"
    desc = u"normalize id attributes."

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for node in xml_root.xpath('.//*[@id]'):
            current_id = node.get('id')
            if VALID_ID_RGX.match(current_id) is None:
                normalized_id = normalize_xml_id(current_id)
                if normalized_id:
                    node.attrib['id'] = normalized_id
                else:
                    node.attrib.pop('id')
                count += 1
                if log_details:
                    log_data.append(log_element(node))
            node.tag = node.tag.lower()
        if count > 0:
            logger.warning(u"{0:d} id attributes got normalized",
                           u"".format(count))
            if log_details:
                logger.warning(u"The following elements have been corrected:",
                               u"\n".join(log_data))
        return xml_root



# -*- coding: utf-8 -*-

from os import path as osp
from lxml import etree
from glamconv.transformer.actions import ValidateAction


DTD_DIR = osp.join(osp.abspath(osp.dirname(__file__)), "ead-2002-dtd")
EAD_DTD = etree.DTD(osp.join(DTD_DIR, "ead.dtd"))

SCHEMA_DIR = osp.join(osp.abspath(osp.dirname(__file__)), "ape-ead-schema")
EAD_SCHEMA_XML = etree.parse(osp.join(SCHEMA_DIR, "apeEAD.xsd"))
EAD_SCHEMA = etree.XMLSchema(EAD_SCHEMA_XML)


class Ead2002Validator(ValidateAction):
    uid = u"ead-2002-validator"
    name = u"Validating XML data in EAD-2002 format"
    desc = (u"This action validates XML data against the EAD-2002 DTD.")

    def _execute(self, xml_root, logger, log_details):
        result = EAD_DTD.validate(xml_root)
        if result is False:
            error_list = [u"Line {0}: {1}".format(err.line, err.message)
                          for err in EAD_DTD.error_log]
            err_num = len(error_list)
            if err_num > 100:
                error_list = error_list[:100]
                error_list.append(u"... and {0:d} more errors! ..."
                                  u"".format((err_num-100)))
            logger.error(
                u"The XML data doesn't comply with the constraints defined in "
                u"the EAD-2002 standard (DTD). The following validation "
                u"errors have been found:",
                u"\n".join(error_list))
        else:
            logger.info(u"The XML data complies with the EAD-2002 standard "
                        u"(DTD).")
        return result


class ApeEadValidator(ValidateAction):
    uid = u"ape-ead-validator"
    name = u"Validating XML data in Ape-EAD format"
    desc = (u"This action validates XML data against the Ae-EAD XML Schema.")

    def _execute(self, xml_root, logger, log_details):
        result = EAD_SCHEMA.validate(xml_root)
        if result is False:
            error_list = [u"Line {0}: {1}".format(err.line, err.message)
                          for err in EAD_SCHEMA.error_log]
            err_num = len(error_list)
            if err_num > 100:
                error_list = error_list[:100]
                error_list.append(u"... and {0:d} more errors! ..."
                                  u"".format((err_num-100)))
            logger.error(
                u"The XML data doesn't comply with the constraints defined in "
                u"the Ape-EAD standard (XML Schema). The following validation "
                u"errors have been found:",
                u"\n".join(error_list))
        else:
            logger.info(u"The XML data complies with the Ape-EAD standard "
                        u"(XML Schema).")
        return result



# -*- coding: utf-8 -*-
"""
Sub-package containing the formats and the actions related to EAD (Encoded
Archival Description).

EAD exists in two main formats : EAD-2002 and Ape-EAD. The various actions
defined here can be used to convert files in EAD-2002 into Ape-EAD that has
more constraints. Both formats rely on XML files.
"""

import os.path as osp
import json

from .registration import register  # noqa


def ead2_to_ape_default_settings():
    """load default ead2 -> ape transformation settings"""
    settings_filepath = osp.join(osp.dirname(__file__),
                                 'default-ead2-to-ape-settings.json')
    with open(settings_filepath) as inputf:
        return json.load(inputf)



# -*- coding: utf-8 -*-
"""
Module containing actions for processing the links (URLs and internal links)
inside the XML elements.
"""

try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

from six import text_type

from lxml import etree

from glamconv.ead.utils import NS, split_qname, log_element
from glamconv.ead.formats import EAD_2002, EAD_APE
from glamconv.transformer.actions import TransformAction
from glamconv.transformer.parameters import SingleParameter, CouplesParameter


class InternalLinkTransformer(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"internal-link-transformer"
    name = u"Transforming <ref> and <ptr> into <extref> and <extptr>"
    category = u"Links & Refs"
    desc = (u"In Ape-EAD, the elements describing an internal link (<ptr> and "
            u"<ref>) doesn't exist anymore. This action transforms the <ref> "
            u"elements into <extref> elements and the <ptr> elements into "
            u"<extptr> elements. The URL of the new link will have the "
            u"following form: \"#internal-id\".")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for elt in xml_root.xpath('.//ref|.//ptr'):
            _, name = split_qname(elt.tag)
            elt.tag = u"ext{0}".format(name)
            if u"target" in elt.attrib:
                elt.set(u"href", u"#{0}".format(elt.attrib.pop("target")))
            count += 1
            if log_details:
                log_data.append(
                    log_element(elt, attributes=(u"href", u"xlink:href")))
        if count > 0:
            logger.warning(
                u"Transforming elements containing an internal link into "
                u"elements containing an external link. {0:d} elements "
                u"have been modified.".format(count))
            if log_details:
                logger.warning(u"The following elements have been modified:",
                               u"\n".join(log_data))
        return xml_root


class ExternalLinkTransformer(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"external-link-transformer"
    name = u"Transforming <archref>, <bibref>, <extptr>, <extref>"
    category = u"Links & Refs"
    desc = (u"This action transforms the <archref>, <bibref>, <extptr> and "
            u"<extref> elements into the legit element for describing the "
            u"external link in the current parent element (e.g. <bibref> in "
            u"<bibliography>, <extref> in <p>).")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        names = (u"archref", u"bibref", u"extptr", u"extref",)
        legit_elt = {u"bibliography": u"bibref", u"unitid": u"extptr",
                     u"descrules": u"extref", u"item": u"extref",
                     u"p": u"extref", u"repository": u"extref"}
        xpath_req = u" | ".join([u'.//{0}'.format(eltname)
                                 for eltname in names])
        for elt in xml_root.xpath(xpath_req):
            parent = elt.getparent()
            newname = legit_elt.get(split_qname(parent.tag)[1], None)
            if newname is None:
                continue
            _, oldname = split_qname(elt.tag)
            if newname == oldname:
                continue
            elt.tag = newname
            count += 1
            lnk = (elt.get(u"href")
                   or elt.get(u"{{{xlink}}}href".format(**NS), u""))
            if newname == "extptr" and len(elt.text or u"") > 0:
                if log_details:
                    log_data.append(
                        log_element(
                            elt, msg=u"    Text content will be deleted",
                            attributes=(u"href", u"xlink:href"),
                            text_content=True))
                elt.text = None
            elif oldname == "extptr" and len(lnk) > 0:
                elt.text = lnk
                if log_details:
                    log_data.append(
                        log_element(
                            elt, msg=u"    href will be added as text content",
                            attributes=(u"href", u"xlink:href")))
            elif log_details:
                log_data.append(
                    log_element(elt, attributes=(u"href", u"xlink:href"),
                                text_content=True))
        if count > 0:
            logger.warning(
                u"Transforming external link elements into the element that "
                u"can occur inside parent element. {0:d} elements have "
                u"been modified.".format(count))
            if log_details:
                logger.warning(u"The following elements have been modified:",
                               u"\n".join(log_data))
        return xml_root


class DaodescEraser(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"daodesc-eraser"
    name = u"Deleting <daodesc> from <dao> and <daogrp>"
    category = u"Links & Refs"
    desc = (u"<daodesc> elements don't exist anymore in Ape-EAD. This "
            u"action deletes the <daodesc> that were defined in <dao> or "
            u"<daogrp> elements.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for elt in xml_root.xpath('.//daodesc'):
            if log_details:
                log_data.append(log_element(elt, text_content=True))
            elt.getparent().remove(elt)
            count += 1
        if count > 0:
            logger.warning(u"{0:d} <daodesc> elements have been deleted."
                           u"".format(count))
            if log_details:
                logger.warning(u"The following elements have been deleted:",
                               u"\n".join(log_data))
        return xml_root


class DaoContentEraser(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"dao-content-eraser"
    name = u"Deleting content in <dao>"
    category = u"Links & Refs"
    desc = (u"In Ape-EAD, <dao> elements must be empty (no text, not even "
            u"spaces). This action deletes the content that might remain in "
            u"<dao> elements.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for elt in xml_root.xpath('.//dao[text() != "" or node()]'):
            elt.text = None
            for child in elt:
                elt.remove(child)
            count += 1
            if log_details:
                log_data.append(
                    log_element(elt, attributes=(u"href", u"xlink:href")))
        if count > 0:
            logger.warning(
                u"Inner content has been deleted from {0:d} <dao> elements."
                u"".format(count))
            if log_details:
                logger.warning(u"The following <dao> elements have been "
                               u"emptied.", u"\n".join(log_data))
        return xml_root


class DaogrpTransformer(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"daogrp-transformer"
    name = u"Transforming <daogrp> into <dao>"
    category = u"Links & Refs"
    desc = (u"Ape-EAD doesn't have any element for describing the extended "
            u"links and their various parts (locator, arc, resource). The "
            u"only possible links are simple links. Instead of simply "
            u"deleting the <daogrp> elements, this action retrieves the "
            u"<daoloc> that often describe some resources (images, etc.) "
            u"and transforms them into <dao> elements outside the <daogrp>. "
            u"The other elements in the <daogrp> (resources, arcs, etc.) are "
            u"deleted.")

    def _execute(self, xml_root, logger, log_details):
        count1, count2 = 0, 0
        if log_details:
            log_data1 = []
            log_data2 = []
        for daogrp in xml_root.xpath(u'.//daogrp'):
            for daoloc in daogrp.xpath(u'daoloc'):
                daoloc.tag = u"dao"
                label = daoloc.attrib.pop(u"label", u"").strip()
                if label != u"" and u"role" not in daoloc.attrib:
                    daoloc.set(u"role", label)
                daogrp.addprevious(daoloc)
                count1 += 1
                if log_details:
                    log_data1.append(
                        log_element(daoloc,
                                    attributes=(u"href", u"xlink:href")))
            if len(daogrp) > 0:
                count2 += 1
                if log_details:
                    log_data2.append(log_element(daogrp))
            daogrp.getparent().remove(daogrp)
        if count1 > 0:
            logger.warning(
                u"{0:d} <dao> elements have been defined from <daoloc> "
                u"found in <daogrp> elements.".format(count1))
            if log_details:
                logger.warning(u"The following <dao> elements have been "
                               u"defined:", u"\n".join(log_data1))
        if count2 > 0:
            logger.warning(
                u"{0:d} <daogrp> elements that contained data other "
                u"than <daoloc> have been deleted.".format(count2))
            if log_details:
                logger.warning(u"The following non-empty <daogrp> elements "
                               u"have been deleted:", u"\n".join(log_data2))
        return xml_root


class ArchdescCDaoMover(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"archdesc-c-dao-mover"
    name = u"Moving <dao> inside <c> or <archdesc> into <did>"
    category = u"Links & Refs"
    desc = (u"In Ape-EAD, the <dao> elements can't be directly inserted into "
            u"a <c> or an <archdesc> element. This action moves the <dao> "
            u"elements from these <c> or <archdesc> into the <did> child "
            u"element they always contain.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for dao in xml_root.xpath(u'.//c/dao | .//archdesc/dao'):
            dids = dao.xpath(u'../did')
            if len(dids) == 0:
                continue
            did = dids[0]
            did.append(dao)
            count += 1
            if log_details:
                log_data.append(
                    log_element(dao, attributes=(u"href", u"xlink:href")))
        if count > 0:
            logger.warning(
                u"{0:d} <dao> elements have been moved from a <c> or an "
                u"<archdesc> into their <did> sibling.".format(count))
            if log_details:
                logger.warning(u"The following <dao> elements have "
                               u"been moved:", u"\n".join(log_data))
        return xml_root


class OddDaoTransformer(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"odd-dao-transformer"
    name = u"Transforming <dao> inside <odd>"
    category = u"Links & Refs"
    desc = (u"In Ape-EAD, the <dao> elements can't exist inside an <odd> "
            u"element. This action transforms these <dao> elements into "
            u"<extref> elements inside a <list>.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for odd in xml_root.xpath(u'.//odd[dao]'):
            lst_elt = etree.SubElement(odd, u'list')
            for dao in odd.xpath(u'dao'):
                dao.tag = u'extref'
                lnk = (dao.get(u"href")
                       or dao.get(u"{{{xlink}}}href".format(**NS), u""))
                title = (dao.get(u"title")
                         or dao.get(u"{{{xlink}}}title".format(**NS), u""))
                if title != u"":
                    dao.text = title
                else:
                    dao.text = lnk
                itm = etree.SubElement(lst_elt, u'item')
                itm.append(dao)
                count += 1
                if log_details:
                    log_data.append(
                        log_element(dao, attributes=(u"href", u"xlink:href")))
        if count > 0:
            logger.warning(u"{0:d} <dao> elements inside an <odd> have been "
                           u"transformed in <extref> element.".format(count))
            if log_details:
                logger.warning(
                    u"Non-legit <dao> elements have been transformed into the "
                    u"following elements:", u"\n".join(log_data))
        return xml_root


class LinkgrpEraser(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"linkgrp-eraser"
    name = u"Erasing the <linkgrp>"
    category = u"Links & Refs"
    desc = (u"Ape-EAD doesn't have any element for describing the extended "
            u"links and their various parts (locator, arc, resource). The "
            u"only possible links are simple links. This action deletes the "
            u"<linkgrp> elements that describe an extended link.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        for elt in xml_root.xpath('.//linkgrp'):
            if log_details:
                log_data.append(log_element(elt))
            elt.getparent().remove(elt)
            count += 1
        if count > 0:
            logger.warning(u"{0:d} <linkgrp> elements have been deleted."
                           u"".format(count))
            if log_details:
                logger.warning(u"The following elements have been deleted:",
                               u"\n".join(log_data))
        return xml_root


class XLinkAttribSetter(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"xlink-setter"
    name = u"Using xlink attributes in links"
    category = u"Links & Refs"
    desc = (u"This action transforms the attributes of <bibref>, <extprt>, "
            u"<extref> and <dao> elements that describe the links into the "
            u"corresponding xlink attributes.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        names = (u"bibref", u"extptr", u"extref", u"dao",)
        xpath_req = u" | ".join([u'.//{0}'.format(eltname)
                                 for eltname in names])
        attrs = (u"actuate", u"arcrole", u"href", u"role", u"show", u"title")
        actuate_conv = {u"onload": u"onLoad", u"onrequest": u"onRequest",
                        u"actuatenone": u"none", u"actuateother": u"other"}
        show_conv = {u"new": u"new", u"embed": u"embed",
                     u"showother": u"other",
                     u"shownone": u"none", u"replace": u"replace"}
        for elt in xml_root.xpath(xpath_req):
            # Link is always a simple link in Ape-EAD
            elt.attrib.pop(u"type", None)
            elt.set(u"{{{xlink}}}type".format(**NS), u"simple")
            # Set xlink attributes
            num = 0
            for attname in elt.attrib:
                namespace, name = split_qname(attname)
                if namespace == NS[u"xlink"] or name not in attrs:
                    continue
                num = 1
                xlink_name = u"{{{0}}}{1}".format(NS[u"xlink"], name)
                # Delete attribute but keep its value
                attvalue = elt.attrib.pop(attname)
                if namespace != NS[u"xlink"] and xlink_name in elt.attrib:
                    # If already have a xlink attribute with the same local
                    # name, keep this xlink attribute and forget the current
                    # one
                    continue
                if name == "actuate":
                    attvalue = actuate_conv.get(attvalue)
                elif name == "show":
                    attvalue = show_conv.get(attvalue)
                if attvalue is not None:
                    elt.set(xlink_name, attvalue)
            count += num
            if log_details and num > 0:
                log_data.append(log_element(elt, attributes=(u"xlink:href",)))
        if count > 0:
            logger.warning(
                u"{0:d} elements describing a link have had their attributes "
                u"transformed into xlink attributes.".format(count))
            if log_details:
                logger.warning(u"The following elements have been modified:",
                               u"\n".join(log_data))
        return xml_root


class DaoCharacterReplacer(TransformAction):
    applicable_for = (EAD_2002, EAD_APE)
    uid = u"dao-character-replacer"
    name = u"Replacing characters in the <dao> links"
    category = u"Links & Refs"
    desc = (
        u"This action replaces an expression with another one inside the "
        u"resource link (xlink:href) defined in the <dao> elements. Please "
        u"be sure to execute this action after the link attributes have "
        u"been transformed in xlink attributes.")
    params_def = (
        SingleParameter(
            u"old", u"Old",
            u"Expression to be found in the link and to be replaced by "
            u"the new one", u"Text", text_type, u""),
        SingleParameter(
            u"new", u"New",
            u"Expression replacing the old one in the link",
            u"Text", text_type, u"")
    )

    def _execute(self, xml_root, logger, log_details, old, new):
        count = 0
        if log_details:
            log_data = []
        for dao in xml_root.xpath(u'.//dao'):
            old_href = dao.get(u"{{{xlink}}}href".format(**NS))
            if old in old_href:
                new_href = old_href.replace(old, new)
                dao.set(u"{{{xlink}}}href".format(**NS), new_href)
                count += 1
                if log_details:
                    msg = (u"    Old href : {0}\n    New href : {1}"
                           u"".format(old_href, new_href))
                    log_data.append(log_element(dao, msg=msg))
        if count > 0:
            logger.warning(
                u"{0:d} <dao> elements have had their 'xlink:href' attribute "
                u"modified.".format(count))
            if log_details:
                logger.warning(u"The following elements have been modified:",
                               u"\n".join(log_data))
        return xml_root


class DaoAbsoluteUrlBuilder(TransformAction):
    applicable_for = (EAD_2002, EAD_APE)
    uid = u"dao-absolute-url-builder"
    name = u"Building absolute URLs in the <dao> links"
    category = u"Links & Refs"
    desc = (
        u"Builds an absolute URL from the data found inside the xlink:href "
        u"attribute of the <dao> elements. The URL contains an address "
        u"(e.g. http://www.exemple.com/data/{href}) and parameters "
        u"defined as key/value couples (e.g. height: 128, width: 128, "
        "id: {href}). In one of the other, the {href} expression can "
        u"be used to insert the content of the href attribute of the "
        u"<dao> element (cf. previous examples). Please be sure to execute "
        u"this action after the link attributes have been transformed in "
        u"xlink attributes.")
    params_def = (
        SingleParameter(
            u"url_address", u"Address",
            u"URL address. The expression {href} can be used to insert the "
            u"content of the href attribute of the <dao> element.", u"Text",
            text_type, u""),
        CouplesParameter(
            u"url_params", u"Parameters",
            u"URL parameters defined as a sequence of couples key/value. In "
            u"these values, the expression {href} can be used to insert the "
            u"content of the href attribute of the <dao> element.",
            u"List of couples (text, text)", text_type, text_type, [])
    )

    def _execute(self, xml_root, logger, log_details, url_address, url_params):
        count = 0
        if log_details:
            log_data = []
        for dao in xml_root.xpath(u'.//dao'):
            old_href = dao.get(u"{{{xlink}}}href".format(**NS))
            if u"{href}" in url_address:
                new_href = url_address.format(href=old_href)
            else:
                new_href = url_address
            params = []
            for key, value in url_params:
                if u"{href}" in value:
                    params.append((key, value.format(href=old_href)))
                else:
                    params.append((key, value))
            if len(params) > 0:
                new_href += "?{0}".format(urlencode(params))
            dao.set(u"{{{xlink}}}href".format(**NS), new_href)
            count += 1
            if log_details:
                msg = (u"    Old href : {0}\n    New href : {1}"
                       u"".format(old_href, new_href))
                log_data.append(log_element(dao, msg=msg))
        if count > 0:
            logger.warning(
                u"{0:d} <dao> elements have had an absolute URL built "
                u"in their 'xlink:href' attribute.".format(count))
            if log_details:
                logger.warning(u"The following elements have been modified:",
                               u"\n".join(log_data))
        return xml_root



#!/usr/bin/env python
# pylint: disable=W0142,W0403,W0404,W0613,W0622,W0622,W0704,R0904,C0103,E0611
#
# You should have received a copy of the GNU Lesser General Public License
# along with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""glamconv setup module"""

import os.path as osp

from setuptools import find_packages, setup


description = u"""set of tools to import, transform and manipulate standard GLAM formats (EAD, Unimarc, etc.)"""

requires = {
    'six': '>= 1.4.0',
    'lxml': None,
}

install_requires = ["{0} {1}".format(d, v and v or "").strip()
                    for d, v in requires.items()]

with open(osp.join(osp.dirname(__file__), 'README')) as f:
    long_description = f.read()


setup(
    name='glamconv',
    version='0.1.7',
    license='LGPL',
    description=description,
    long_description=description,
    author='Logilab S.A. (Paris, FRANCE)',
    author_email='contact@logilab.fr',
    url='http://www.logilab.org/project/glamconv',
    classifiers=[
        'Programming Language :: Python',
    ],
    packages=find_packages(exclude=['test']),
    install_requires=install_requires,
    include_package_data=True,
    # entry_points={},  # XXX TODO
    zip_safe=False,
)



# -*- coding: utf-8 -*-
"""
Module containing actions for converting, moving or correcting XML elements
that occur inside the header or the front-matter.
"""

from six import text_type

from lxml import etree

from glamconv.ead.utils import log_element, insert_child_at_element_beginning
from glamconv.ead.formats import EAD_2002
from glamconv.transformer.actions import TransformAction
from glamconv.transformer.parameters import SingleParameter


class IdentifiersDefiner(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"identifiers-definer"
    name = u"Defining identifiers"
    category = u"Header"
    desc = (u"This action defines the identifiers inside the <eadid> element, "
            u"if they are not already defined.")
    params_def = (
        SingleParameter(
            u"country_code", u"Country code",
            u"Code of the country where dwells the organization that "
            u"publishes the EAD. Only some codes are allowed. They are in "
            u"upper case and contain two letters.", u"Text", text_type, u""),
        SingleParameter(
            u"agency_code", u"Agency code",
            u"Code of the organization that publishes the EAD. The codes are "
            u"actually in upper case and begin with the country code.",
            u"Text", text_type, u""),
        SingleParameter(
            u"document_id", u"Document identifier",
            u"Identifier of the document. The identifier usually starts with "
            u"the agency code.", u"Text", text_type, u""),
        SingleParameter(
            u"favour_xml_document_id",
            u"Favour XML data for document identifier",
            u"If the 'identifier' attribute of <eadid> element is not "
            u"defined, it is possible either to gather data from the XML to "
            u"build it (typically the text inside the <eadid> element) or to "
            u"use the value given in the \"Document indentifier\" parameter "
            u"of this action. When this parameter is set, this action will "
            u"first try to use the XML data to build the document identifier "
            u"and will only use the parameter value if no XML data can be "
            u"found. When this parameter is unset, this action will always "
            u"use the parameter value without using the XML data/.",
            u"Boolean", bool, False),
    )

    def _execute(self, xml_root, logger, log_details, country_code,
                 agency_code, document_id, favour_xml_document_id):
        for eadid in xml_root.xpath('.//eadid'):
            if country_code and country_code.upper() != 'FR':
                logger.info('ignoring specified country code %s',
                            country_code.upper())
            eadid.set(u"countrycode", 'FR')
            if agency_code != u"":
                eadid.set(u"mainagencycode", agency_code)
                logger.info(
                    u"Setting the 'mainagencycode' attribute in the <eadid> "
                    u"element", u"    New value: {0}".format(agency_code))
            xml_built_id = (eadid.text or u"").strip()
            if document_id != u"" \
               and (not(favour_xml_document_id) or len(xml_built_id) == 0):
                eadid.set(u"identifier", document_id)
                logger.info(
                    u"Setting the 'identifier' attribute in the <eadid> "
                    u"element from the document_id parameter",
                    u"    New value: {0}".format(document_id))
            elif (favour_xml_document_id and len(xml_built_id) > 0
                  and u"identifier" not in eadid.attrib):
                eadid.set(u"identifier", xml_built_id)
                logger.info(
                    u"Setting the 'identifier' attribute in the <eadid> "
                    u"element from the XML data read in <eadid>",
                    u"    New value: {0}".format(xml_built_id))
        return xml_root


def _add_content_to_odd_elt(xml_root, elements, title, log_data):
    count = 0
    odd_lst = xml_root.xpath(u".//archdesc[1]/odd")
    if len(odd_lst) > 0:
        odd = odd_lst[0]
    else:
        odd = etree.Element(u"odd")
        dsc_lst = xml_root.xpath(u".//archdesc[1]/dsc")
        if len(dsc_lst) > 0:
            dsc_lst[0].addprevious(odd)
        else:
            xml_root.xpath(u".//archdesc[1]")[0].append(odd)
    if title != u"":
        p = etree.SubElement(odd, u'p')
        emph = etree.SubElement(p, u'emph')
        emph.set(u"render", u"bold")
        emph.text = title
    for elt in elements:
        odd.append(elt)
        if log_data is not None:
            log_data.append(log_element(elt, text_content=True))
        count += 1
    return count


class EditionstmtConverter(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"editionstmt-converter"
    name = u"Moving the content of <editionstmt> into <odd>"
    category = u"Header"
    desc = (u"The <editionstmt> element doesn't exist anymore. This action "
            u"moves its content into the <processinfo> element of the "
            u"<archdesc> element. The couples <edition> / <p> in <editionstmt>"
            u"are transformed into list items before being moved in <odd>.")
    params_def = (
        SingleParameter(
            u"title", u"Title",
            u"Title inserted in the <odd> element before the <editionstmt> "
            u"content.", u"Text", text_type, u"Editions"),
    )

    def _execute(self, xml_root, logger, log_details, title):
        count = 0
        log_data = None
        if log_details:
            log_data = []
        list_elt = etree.Element(u"list")
        for edt in xml_root.xpath(u'.//editionstmt/edition'):
            nxt = edt.xpath(u"following-sibling::*")
            if len(nxt) == 0 or nxt[0].tag != u"p":
                itm = etree.Element(u"item")
            else:
                itm = nxt[0]
                itm.tag = u"item"
            list_elt.append(itm)
            emph = etree.Element(u"emph", render=u"bold")
            emph.text = edt.text.strip()
            emph.tail = u": "
            insert_child_at_element_beginning(itm, emph)
        if len(list_elt) > 0:
            count += _add_content_to_odd_elt(xml_root, [list_elt], title,
                                             log_data)
            logger.warning(
                u"Moving the content of <editionstmt> elements into the <odd> "
                u"element in <archdesc>. {0:d} elements have been inserted in "
                u"<odd>".format(count))
            if log_details:
                logger.warning(u"The following elements have been inserted in "
                               u"<odd>:", u"\n".join(log_data))
        for stmt in xml_root.xpath(u'.//editionstmt'):
            stmt.getparent().remove(stmt)
        return xml_root


class NotestmtConverter(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"notestmt-converter"
    name = u"Moving the content of <notestmt> into <odd>"
    category = u"Header"
    desc = (u"The <notestmt> element doesn't exist anymore. This action moves "
            u"its content into the <odd> element in <archdesc>. The <note> "
            u"elements inside <notestmt> are inserted into <p> and then moved "
            u"inside <odd> (as Ape-EAD doesn't allow the notes inside odd).")
    params_def = (
        SingleParameter(
            u"title", u"Title",
            u"Title inserted in the <odd> element before the <notestmt> "
            u"content.", u"Text", text_type, u"Notes"),
    )

    def _execute(self, xml_root, logger, log_details, title):
        count = 0
        log_data = None
        if log_details:
            log_data = []
        elts_to_move = []
        for note in xml_root.xpath(u'.//notestmt/note'):
            p = etree.Element(u"p")
            p.append(note)
            elts_to_move.append(p)
        if len(elts_to_move) > 0:
            count += _add_content_to_odd_elt(xml_root, elts_to_move, title,
                                             log_data)
            logger.warning(
                u"Moving the content of <notestmt> elements into the <odd> "
                u"element in <archdesc>. {0:d} elements have been inserted in "
                u"<odd>".format(count))
            if log_details:
                logger.warning(u"The following elements have been inserted in "
                               u"<odd>:", u"\n".join(log_data))
        for stmt in xml_root.xpath(u'.//notestmt'):
            stmt.getparent().remove(stmt)
        return xml_root


def _div_collapser(div_elt, elts_to_move, level=0):
    head = div_elt.find(u"head")
    if head is not None:
        head.text = (u"> "*level) + (head.text or u"")
        head.tag = u"p"
        if len(head) == 0:
            emph = etree.Element(u"emph", render=u"italic")
            emph.text = head.text
            head.text = None
            head.append(emph)
        elts_to_move.append(head)
    for child in div_elt.xpath(u"*[not(self::head)]"):
        if child.tag == u"div":
            _div_collapser(child, elts_to_move, level+1)
        else:
            elts_to_move.append(child)


class FrontmatterConverter(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"frontmatter-converter"
    name = u"Moving the content of <frontmatter> into <odd>"
    category = u"Header"
    desc = (
        u"The <frontmatter> element doesn't exist anymore in Ape-EAD. This "
        u"action moves its content into the <odd> element in <archdesc> "
        u"element. Be aware that the <titlepage> sub-element is not moved "
        u"and thus is deleted.")
    params_def = (
        SingleParameter(
            u"title", u"Title",
            u"Title inserted in the <odd> element before the <frontmatter> "
            u"content.", u"Text", text_type, u"Prolegomena"),
    )

    def _execute(self, xml_root, logger, log_details, title):
        count = 0
        log_data = None
        if log_details:
            log_data = []
        elts_to_move = []
        for div in xml_root.xpath(u'.//frontmatter/div'):
            _div_collapser(div, elts_to_move)
        if len(elts_to_move) > 0:
            count += _add_content_to_odd_elt(xml_root, elts_to_move, title,
                                             log_data)
            logger.warning(
                u"Moving the content of <frontmatter> elements into the <odd> "
                u"element in <archdesc> (except <titlepage>). {0:d} elements "
                u"have been inserted in <odd>".format(count))
            if log_details:
                logger.warning(u"The following elements have been inserted in "
                               u"<odd>:", u"\n".join(log_data))
        if log_details:
            for ttpg in xml_root.xpath(u'.//frontmatter/titlepage'):
                logger.warning(
                    u"Deleting <titlepage> element inside <frontmatter>. The "
                    u"following sub-elements have been deleted:",
                    log_element(ttpg, text_content=True))
        for ftmtt in xml_root.xpath(u'.//frontmatter'):
            ftmtt.getparent().remove(ftmtt)
        return xml_root


class SponsorConverter(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"sponsor-converter"
    name = u"Moving <sponsor> from <titlestmt> into <odd>"
    category = u"Header"
    desc = (u"In Ape-EAD, the <sponsor> element that could occur in "
            u"<titlestmt> doesn't exist anymore. This action transforms it "
            u"into a paragraph (<p>) and moves it into the <odd> element in "
            u"<archdesc> element.")
    params_def = (
        SingleParameter(
            u"title", u"Title",
            u"Title inserted in the <odd> element before the <sponsor> "
            u"content.", u"Text", text_type, u"Sponsor: "),
    )

    def _execute(self, xml_root, logger, log_details, title):
        count = 0
        log_data = None
        if log_details:
            log_data = []
        elts_to_move = []
        for spsr in xml_root.xpath(u'.//titlestmt/sponsor'):
            spsr.tag = u'p'
            emph = etree.Element(u'emph', render=u"bold")
            emph.text = title
            emph.tail = u" "
            insert_child_at_element_beginning(spsr, emph)
            elts_to_move.append(spsr)
        if len(elts_to_move) > 0:
            count += _add_content_to_odd_elt(xml_root, elts_to_move, title,
                                             log_data)
            logger.warning(
                u"{0:d} <sponsor> elements have been moved into the "
                u"<odd> element in <archdesc>.".format(count))
            if log_details:
                logger.warning(u"The following elements have been moved:",
                               u"\n".join(log_data))
        return xml_root


class StmtPNumEraser(TransformAction):
    applicable_for = (EAD_2002,)
    uid = u"stmt-p-num-eraser"
    name = u"Erasing <p>, <num> from <publicationstmt>, <seriesstmt> "
    category = u"Header"
    desc = (u"In Ape-EAD, <p> and <num> elements cannot occur inside "
            u"<publicationstmt> or <seriestmt> elements. This action deletes "
            u"these non-legit child elements.")

    def _execute(self, xml_root, logger, log_details):
        count = 0
        if log_details:
            log_data = []
        xpath_req = (u'.//publicationstmt/p | .//publicationstmt/num | '
                     u'.//seriesstmt/p | .//seriesstmt/num')
        for elt in xml_root.xpath(xpath_req):
            count += 1
            if log_details:
                log_data.append(log_element(elt, text_content=True))
            elt.getparent().remove(elt)
        if count > 0:
            logger.warning(
                u"{0:d} <p> and <num> elements have been deleted "
                u"from <publicationstmt> and <seriesstmt>.".format(count))
            if log_details:
                logger.warning(u"The following elements have been suppressed:",
                               u"\n".join(log_data))
        return xml_root
